from functools import reduce
from typing import List, Optional, Tuple, Union, overload

from PySide6.QtCore import QObject, QPoint, QRect, QSize, Signal, SignalInstance

from foundry.game.File import ROM
from foundry.game.gfx.objects.EnemyItem import EnemyObject
from foundry.game.gfx.objects.EnemyItemFactory import EnemyItemFactory
from foundry.game.gfx.objects.Jump import Jump
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.gfx.objects.LevelObjectFactory import LevelObjectFactory
from foundry.game.level import LevelByteData
from foundry.game.level.LevelLike import LevelLike
from foundry.game.level.util import get_worlds, load_level_offsets
from foundry.game.ObjectSet import ObjectSet
from foundry.smb3parse.constants import (
    BASE_OFFSET,
    TILESET_LEVEL_OFFSET,
    Level_TilesetIdx_ByTileset,
)
from foundry.smb3parse.levels.level_header import LevelHeader

LEVEL_POINTER_OFFSET = Level_TilesetIdx_ByTileset

ENEMY_SIZE = 3

TIME_INF = -1

LEVEL_DEFAULT_HEIGHT = 27
LEVEL_DEFAULT_WIDTH = 16


def get_level_name_suggestion(level_address: int) -> str:
    for level in Level.offsets:
        if level.generator_pointer == level_address:
            name = level.display_information.name
            if name is not None:
                return name
            return "Unspecified"
    else:
        return "Unknown"


class LevelSignaller(QObject):
    data_changed: SignalInstance = Signal()
    jumps_changed: SignalInstance = Signal()


class Level(LevelLike):
    MIN_LENGTH = 0x10

    offsets = load_level_offsets()
    sorted_offsets = sorted(offsets, key=lambda level: level.generator_pointer)

    WORLDS = get_worlds(offsets)

    HEADER_LENGTH = 9  # bytes

    def __init__(
        self, level_name: str = "", layout_address: int = 0, enemy_data_offset: int = 0, object_set_number: int = 1
    ):
        super(Level, self).__init__(object_set_number, layout_address)

        self._signal_emitter = LevelSignaller()

        self.changed = False
        """Whether the current level was modified since it was loaded/last saved."""

        self.object_set = ObjectSet(object_set_number)

        self.name = level_name

        self.header_offset = layout_address
        self.object_offset = self.header_offset + Level.HEADER_LENGTH
        self.enemy_offset = enemy_data_offset

        self.objects: List[LevelObject] = []
        self.header_bytes: bytearray = bytearray()
        self.jumps: List[Jump] = []
        self.enemies: List[EnemyObject] = []

        if self.layout_address == self.enemy_offset == 0:
            # probably loaded to become an m3l
            return

        rom = ROM()

        self.header_bytes = rom.bulk_read(Level.HEADER_LENGTH, self.header_offset)
        self._parse_header()

        object_data = ROM.rom_data[self.object_offset :]
        enemy_data = ROM.rom_data[self.enemy_offset :]

        self._load_level_data(object_data, enemy_data)

    def _load_level_data(self, object_data: bytearray, enemy_data: bytearray, new_level: bool = True):
        self._load_objects(object_data)
        self._load_enemies(enemy_data)

        if new_level:
            self._update_level_size()
            self.data_changed.emit()

    @property
    def fully_loaded(self):
        """Whether this object represents a fully loaded Level, meaning it was either loaded from a ROM or from an m3l
        file. If this is false, it is probably just a place holder to use either from_bytes or from_m3l later."""
        # objects, enemies and jumps could be empty, but there are always 9 header bytes, when a level is loaded
        return bool(self.header_bytes)

    @property
    def attached_to_rom(self):
        """Whether the current level has a place in the ROM yet. If not this level is likely a m3l file."""
        return not (self.header_offset == self.enemy_offset == 0)

    @attached_to_rom.setter
    def attached_to_rom(self, should_be_attached):
        if not should_be_attached:
            self.header_offset = self.enemy_offset = 0
        else:
            raise ValueError("You cannot manually attach a ROM, use attach_to_rom() instead.")

    @property
    def width(self):
        return self.size[0]

    @property
    def height(self):
        return self.size[1]

    @property
    def data_changed(self):
        return self._signal_emitter.data_changed

    @property
    def jumps_changed(self):
        return self._signal_emitter.jumps_changed

    def reload(self):
        (_, header_and_object_data), (_, enemy_data) = self.to_bytes()

        self.header_bytes = header_and_object_data[: Level.HEADER_LENGTH]

        object_data = header_and_object_data[Level.HEADER_LENGTH :]

        self._parse_header()
        self._load_level_data(object_data, enemy_data, new_level=False)

        self.data_changed.emit()

    def current_object_size(self):
        return reduce(
            lambda count, element: count + len(element), [obj.to_bytes() for obj in self.objects + self.jumps], 0
        )

    def current_enemies_size(self):
        return reduce(lambda count, element: count + len(element), [obj.to_bytes() for obj in self.enemies], 0)

    def _parse_header(self):
        self.header = LevelHeader(self.header_bytes, self.object_set_number)

        self.object_factory = LevelObjectFactory(
            self.object_set_number,
            self.header.graphic_set_index,
            self.header.object_palette_index,
            self.objects,
            bool(self.header.is_vertical),
        )
        self.enemy_item_factory = EnemyItemFactory(self.object_set_number, self.header.enemy_palette_index)

        self.size = self.header.width, self.header.height

        self.data_changed.emit()

    def _load_enemies(self, data: bytearray):
        self.enemies.clear()

        def data_left(_data: bytearray):
            # the commented out code seems to hold for the stock ROM, but if the ROM was already edited with another
            # editor, it might not, since they only wrote the 0xFF to end the enemy data

            return _data and not _data[0] == 0xFF  # and _data[1] in [0x00, 0x01]

        enemy_data, data = data[0:ENEMY_SIZE], data[ENEMY_SIZE:]

        while data_left(enemy_data):
            enemy = self.enemy_item_factory.from_data(enemy_data, 0)

            self.enemies.append(enemy)

            enemy_data, data = data[0:ENEMY_SIZE], data[ENEMY_SIZE:]

    def _load_objects(self, data: bytearray):
        self.objects.clear()
        self.jumps.clear()

        if not data or data[0] == 0xFF:
            return

        while True:
            obj_data, data = data[0:3], data[3:]

            domain = (obj_data[0] & 0b1110_0000) >> 5

            obj_id = obj_data[2]
            has_length_byte = self.object_set.get_object_byte_length(domain, obj_id) == 4

            if has_length_byte:
                fourth_byte, data = data[0], data[1:]
                obj_data.append(fourth_byte)

            level_object = self.object_factory.from_data(obj_data, len(self.objects))

            if isinstance(level_object, LevelObject):
                self.objects.append(level_object)
            elif isinstance(level_object, Jump):
                self.jumps.append(level_object)

            if data[0] == 0xFF:
                break

    def _update_level_size(self):
        self.object_size_on_disk = self.current_object_size()
        self.enemy_size_on_disk = self.current_enemies_size()

    @property
    def size_on_disk(self):
        return self.object_size_on_disk + self.enemy_size_on_disk

    def get_rect(self, block_length: int = 1):
        width, height = self.size

        return QRect(QPoint(0, 0), QSize(width, height) * block_length)

    def attach_to_rom(self, header_offset: int, enemy_item_offset: int):
        if 0x0 in [header_offset, enemy_item_offset]:
            raise ValueError("You cannot save level or enemy data to the beginning of the ROM (address 0x0).")

        self.header_offset = header_offset
        self.object_offset = self.header_offset + Level.HEADER_LENGTH
        self.enemy_offset = enemy_item_offset

    def was_saved(self):
        self._update_level_size()

    @property
    def objects_end(self):
        return self.header_offset + Level.HEADER_LENGTH + self.current_object_size() + len(b"\xFF")  # the delimiter

    @property
    def enemies_end(self):
        return self.enemy_offset + self.current_enemies_size() + len(b"\xFF\x00")  # the delimiter

    @property
    def next_area_objects(self):
        return self.header.jump_level_address

    @next_area_objects.setter
    def next_area_objects(self, value):
        if value == self.header.jump_level_address:
            return

        value -= LEVEL_POINTER_OFFSET + TILESET_LEVEL_OFFSET[self.header.jump_object_set_number]

        self.header_bytes[0] = 0x00FF & value
        self.header_bytes[1] = value >> 8

        self._parse_header()

    @property
    def has_next_area(self):
        return self.header_bytes[0] + self.header_bytes[1] != 0

    @property
    def next_area_enemies(self):
        return self.header.jump_enemy_address

    @next_area_enemies.setter
    def next_area_enemies(self, value):
        if value == self.header.jump_enemy_address:
            return

        value -= BASE_OFFSET

        self.header_bytes[2] = 0x00FF & value
        self.header_bytes[3] = value >> 8

        self._parse_header()

    @property
    def start_y_index(self):
        return self.header.start_y_index

    @start_y_index.setter
    def start_y_index(self, index):
        if index == self.header.start_y_index:
            return

        self.header_bytes[4] &= 0b0001_1111
        self.header_bytes[4] |= index << 5

        self._parse_header()

    # bit 4 unused

    @property
    def length(self):
        return self.header.length

    @length.setter
    def length(self, length):
        """
        Sets the length of the level in "screens".

        :param length: amount of screens the level should have
        :return:
        """

        if length == self.header.length:
            return

        # screens are 0 indexed, minimum is 1
        self.header_bytes[4] &= 0b1111_0000
        self.header_bytes[4] |= (length // 0x10) - 1

        self._parse_header()

    # bit 1 unused

    @property
    def start_x_index(self):
        return self.header.start_x_index

    @start_x_index.setter
    def start_x_index(self, index):
        if index == self.header.start_x_index:
            return

        self.header_bytes[5] &= 0b1001_1111
        self.header_bytes[5] |= index << 5

        self._parse_header()

    @property
    def enemy_palette_index(self):
        return self.header.enemy_palette_index

    @enemy_palette_index.setter
    def enemy_palette_index(self, index):
        if index == self.header.enemy_palette_index:
            return

        self.header_bytes[5] &= 0b1110_0111
        self.header_bytes[5] |= index << 3

        self._parse_header()

    @property
    def object_palette_index(self):
        return self.header.object_palette_index

    @object_palette_index.setter
    def object_palette_index(self, index):
        if index == self.header.object_palette_index:
            return

        self.header_bytes[5] &= 0b1111_1000
        self.header_bytes[5] |= index

        self._parse_header()

        self.reload()

    @property
    def pipe_ends_level(self):
        return self.header.pipe_ends_level

    @pipe_ends_level.setter
    def pipe_ends_level(self, truth_value):
        if truth_value == self.header.pipe_ends_level:
            return

        self.header_bytes[6] &= 0b0111_1111
        self.header_bytes[6] |= int(not truth_value) << 7

        self._parse_header()

    @property
    def scroll_type(self):
        return self.header.scroll_type_index

    @scroll_type.setter
    def scroll_type(self, index):
        if index == self.header.scroll_type_index:
            return

        self.header_bytes[6] &= 0b1001_1111
        self.header_bytes[6] |= index << 5

        self._parse_header()

    @property
    def is_vertical(self):
        return bool(self.header.is_vertical)

    @is_vertical.setter
    def is_vertical(self, truth_value):
        if truth_value == self.header.is_vertical:
            return

        self.header_bytes[6] &= 0b1110_1111
        self.header_bytes[6] |= int(truth_value) << 4

        self._parse_header()

    @property
    def next_area_object_set(self):
        return self.header.jump_object_set_number

    @next_area_object_set.setter
    def next_area_object_set(self, index):
        if index == self.header.jump_object_set_number:
            return

        self.header_bytes[6] &= 0b1111_0000
        self.header_bytes[6] |= index

        self._parse_header()

    @property
    def start_action(self):
        return self.header.start_action

    @start_action.setter
    def start_action(self, index):
        if index == self.header.start_action:
            return

        self.header_bytes[7] &= 0b0001_1111
        self.header_bytes[7] |= index << 5

        self._parse_header()

    @property
    def graphic_set(self):
        return self.header.graphic_set_index

    @graphic_set.setter
    def graphic_set(self, index):
        if index == self.header.graphic_set_index:
            return

        self.header_bytes[7] &= 0b1110_0000
        self.header_bytes[7] |= index

        self._parse_header()

        self.reload()

    @property
    def time_index(self):
        return self.header.time_index

    @time_index.setter
    def time_index(self, index):
        if index == self.header.time_index:
            return

        self.header_bytes[8] &= 0b0011_1111
        self.header_bytes[8] |= index << 6

        self._parse_header()

    # bit 3 and 4 unused

    @property
    def music_index(self):
        return self.header.music_index

    @music_index.setter
    def music_index(self, index):
        if index == self.header.music_index:
            return

        self.header_bytes[8] &= 0b1111_0000
        self.header_bytes[8] |= index

        self._parse_header()

    def is_too_big(self):
        return self.too_many_level_objects() or self.too_many_enemies_or_items()

    def too_many_level_objects(self):
        return self.current_object_size() > self.object_size_on_disk

    def too_many_enemies_or_items(self):
        return self.current_enemies_size() > self.enemy_size_on_disk

    def get_all_objects(self) -> List[Union[LevelObject, EnemyObject]]:
        return self.objects + self.enemies

    def get_object_names(self):
        return [obj.name for obj in self.get_all_objects()]

    def object_at(self, x: int, y: int) -> Optional[Union[EnemyObject, LevelObject]]:
        for obj in reversed(self.get_all_objects()):
            if (x, y) in obj:
                return obj
        else:
            return None

    def bring_to_foreground(self, objects: List[Union[LevelObject, EnemyObject]]):
        for obj in objects:
            intersecting_objects = self.get_intersecting_objects(obj)

            object_currently_in_the_foreground: Union[LevelObject, EnemyObject] = intersecting_objects[-1]

            if obj is object_currently_in_the_foreground:
                continue

            if isinstance(obj, LevelObject):
                objects = self.objects
            elif isinstance(obj, EnemyObject):
                objects = self.enemies

            objects.remove(obj)

            index = objects.index(object_currently_in_the_foreground) + 1

            objects.insert(index, obj)

    def bring_to_background(self, level_objects: List[Union[LevelObject, EnemyObject]]):
        for obj in level_objects:
            intersecting_objects = self.get_intersecting_objects(obj)

            object_currently_in_the_background: Union[LevelObject, EnemyObject] = intersecting_objects[0]

            if obj is object_currently_in_the_background:
                continue

            if isinstance(obj, LevelObject):
                objects = self.objects
            elif isinstance(obj, EnemyObject):
                objects = self.enemies
            else:
                raise TypeError()

            objects.remove(obj)

            index = objects.index(object_currently_in_the_background)

            objects.insert(index, obj)

    @overload
    def get_intersecting_objects(self, obj: LevelObject) -> List[LevelObject]:
        ...

    @overload
    def get_intersecting_objects(self, obj: EnemyObject) -> List[EnemyObject]:
        ...

    def get_intersecting_objects(
        self, obj: Union[LevelObject, EnemyObject]
    ) -> Union[List[LevelObject], List[EnemyObject]]:
        """
        Returns all objects of the same type, that overlap the rectangle of the given object, including itself. The
        objects are in the order, that they appear in, in memory, meaning back to front.

        :param obj: The object to check overlaps for.
        :return:
        """
        if isinstance(obj, LevelObject):
            objects_to_check = self.objects
        elif isinstance(obj, EnemyObject):
            objects_to_check = self.enemies
        else:
            raise TypeError()

        intersecting_objects = []

        for other_object in objects_to_check:
            if obj.get_rect().intersects(other_object.get_rect()):
                intersecting_objects.append(other_object)

        return intersecting_objects

    def draw(self, *_):
        pass

    def paste_object_at(self, x: int, y: int, obj: Union[EnemyObject, LevelObject]) -> Union[EnemyObject, LevelObject]:
        if isinstance(obj, EnemyObject):
            return self.add_enemy(obj.obj_index, x, y)

        elif isinstance(obj, LevelObject):
            if obj.is_4byte:
                length: Optional[int] = obj.data[3]
            else:
                length = None

            return self.add_object(obj.domain, obj.obj_index, x, y, length)

    def create_object_at(self, x: int, y: int, domain: int = 0, object_index: int = 0):
        assert isinstance(domain, int)
        assert isinstance(object_index, int)
        assert isinstance(x, int)
        assert isinstance(y, int)
        assert isinstance(object_index, int)

        self.add_object(domain, object_index, x, y, None, len(self.objects))

    def create_enemy_at(self, x: int, y: int):
        # goomba to have something to display
        self.add_enemy(0x72, x, y, len(self.enemies))

    def add_object(
        self, domain: int, object_index: int, x: int, y: int, length: Optional[int], index: int = -1
    ) -> LevelObject:
        assert isinstance(domain, int)
        assert isinstance(object_index, int)
        assert isinstance(x, int)
        assert isinstance(y, int)
        assert isinstance(index, int)

        if index == -1:
            index = len(self.objects)

        obj = self.object_factory.from_properties(domain, object_index, x, y, length, index)
        self.objects.insert(index, obj)

        return obj

    def add_enemy(self, object_index: int, x: int, y: int, index: int = -1) -> EnemyObject:
        if index == -1:
            index = len(self.enemies)
        else:
            index %= len(self.objects)

        enemy = self.enemy_item_factory.from_data([object_index, x, y], -1)

        self.enemies.insert(index, enemy)

        return enemy

    def add_jump(self):
        self.jumps.append(Jump.from_properties(0, 0, 0, 0))

        self.data_changed.emit()

    def remove_jump(self, jump: Jump):
        self.jumps.remove(jump)

        self.data_changed.emit()

    def index_of(self, obj: Union[EnemyObject, LevelObject]) -> int:
        if isinstance(obj, LevelObject):
            return self.objects.index(obj)
        elif isinstance(obj, EnemyObject):
            return len(self.objects) + self.enemies.index(obj)
        else:
            raise TypeError("Given Object was not EnemyObject or LevelObject.")

    def get_object(self, index: int):
        if index < len(self.objects):
            return self.objects[index]
        else:
            return self.enemies[index % len(self.objects)]

    def remove_object(self, obj: Union[EnemyObject, LevelObject]):
        if obj is None:
            return

        if isinstance(obj, LevelObject):
            self.objects.remove(obj)
        elif isinstance(obj, EnemyObject):
            self.enemies.remove(obj)

    def to_m3l(self) -> bytearray:
        world_number = level_number = 1

        m3l_bytes = bytearray()

        m3l_bytes.append(world_number)
        m3l_bytes.append(level_number)
        m3l_bytes.append(self.object_set_number)

        m3l_bytes.extend(self.header_bytes)

        for obj in self.objects + self.jumps:
            m3l_bytes.extend(obj.to_bytes())

        # only write 0xFF, even though the stock ROM would use 0xFF00 or 0xFF01
        # this is done to keep compatibility to older editors
        m3l_bytes.append(0xFF)
        m3l_bytes.append(0x01)

        for enemy in sorted(self.enemies, key=lambda _enemy: _enemy.position.x):
            m3l_bytes.extend(enemy.to_bytes())

        m3l_bytes.append(0xFF)

        return m3l_bytes

    def from_m3l(self, m3l_bytes: bytearray):
        world_number, level_number, self.object_set_number = m3l_bytes[:3]
        self.object_set = ObjectSet(self.object_set_number)

        self.header_offset = self.enemy_offset = 0

        # block signals, so it will only be emitted, once we are fully set up
        self._signal_emitter.blockSignals(True)

        # update the level_object_factory
        self._load_level_data(bytearray(), bytearray(), new_level=False)

        m3l_bytes = m3l_bytes[3:]

        self.header_bytes = m3l_bytes[: Level.HEADER_LENGTH]
        self._parse_header()

        m3l_bytes = m3l_bytes[Level.HEADER_LENGTH :]

        # figure out how many bytes are the objects
        self._load_objects(m3l_bytes)
        object_size = self.current_object_size() + len(b"\xFF")  # delimiter

        object_bytes = m3l_bytes[:object_size]
        enemy_bytes = m3l_bytes[object_size:]

        if len(enemy_bytes) % 3 - len(b"\xFF") == 1:
            # compatibility with workshop
            enemy_bytes = enemy_bytes[1:]

        self._signal_emitter.blockSignals(False)

        self._load_level_data(object_bytes, enemy_bytes)

    def to_bytes(self) -> LevelByteData:
        data = bytearray()

        data.extend(self.header_bytes)

        for obj in self.objects:
            data.extend(obj.to_bytes())

        for jump in self.jumps:
            data.extend(jump.to_bytes())

        data.append(0xFF)

        enemies = bytearray()

        if self.is_vertical:
            enemies_objects = sorted(self.enemies, key=lambda _enemy: _enemy.position.y)
        else:
            enemies_objects = sorted(self.enemies, key=lambda _enemy: _enemy.position.x)

        for enemy in enemies_objects:
            enemies.extend(enemy.to_bytes())

        enemies.append(0xFF)

        return (self.header_offset, data), (self.enemy_offset, enemies)

    def from_bytes(self, object_data: Tuple[int, bytearray], enemy_data: Tuple[int, bytearray], new_level=True):

        self.header_offset, object_bytes = object_data
        self.enemy_offset, enemies = enemy_data

        self.header_bytes = object_bytes[0 : Level.HEADER_LENGTH]
        objects = object_bytes[Level.HEADER_LENGTH :]

        self._parse_header()
        self._load_level_data(objects, enemies, new_level)



from PySide6.QtCore import QSize
from PySide6.QtGui import QMouseEvent, Qt
from PySide6.QtWidgets import QListWidget, QScrollBar, QSizePolicy, QWidget

from foundry.game.level.LevelRef import LevelRef
from foundry.gui.ContextMenu import ContextMenu


class ObjectList(QListWidget):
    def __init__(self, parent: QWidget, level_ref: LevelRef, context_menu: ContextMenu):
        super(ObjectList, self).__init__(parent=parent)

        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
        self.setSizeAdjustPolicy(QListWidget.AdjustToContents)
        self.setSelectionMode(QListWidget.ExtendedSelection)
        scroll_bar = QScrollBar(self)
        self.setVerticalScrollBar(scroll_bar)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setWordWrap(True)

        self.setSelectionMode(self.ExtendedSelection)

        self.level_ref: LevelRef = level_ref
        self.level_ref.data_changed.connect(self.update_content)
        self.context_menu = context_menu

        self.labels = []
        self._on_selection_changed_ongoing = False
        self.itemSelectionChanged.connect(self.on_selection_changed)

        self.setWhatsThis(
            "<b>Object List</b><br/>"
            "This lists all the objects and enemies/items in the level. They appear in the order, "
            "that they are stored in the ROM as, which also decides which objects get drawn "
            "before/behind which.<br/>"
            "Enemies/items are always listed last, since they are also stored separately from the level "
            "objects.<br/><br/>"
            "Note: While Jumps are technically level objects, they are omitted here, since they are "
            "listed in a separate list below."
        )

    def sizeHint(self):
        return QSize(100, 200)

    def mousePressEvent(self, event: QMouseEvent):
        if event.button() == Qt.RightButton:
            self.on_right_down(event)
        else:
            return super(ObjectList, self).mousePressEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent):
        if event.button() == Qt.RightButton:
            self.on_right_up(event)
        else:
            return super(ObjectList, self).mouseReleaseEvent(event)

    def on_right_down(self, event: QMouseEvent):
        item_under_mouse = self.itemAt(event.pos())

        if item_under_mouse is None:
            event.ignore()
            return

        if not item_under_mouse.isSelected():
            self.clearSelection()

            index = self.indexFromItem(item_under_mouse)

            selected_object = self.level_ref.level.get_all_objects()[index.row()]

            self.level_ref.selected_objects = [selected_object]

    def on_right_up(self, event):
        item_under_mouse = self.itemAt(event.pos())

        if item_under_mouse is None:
            event.ignore()
            return

        self.context_menu.as_list_menu().popup(event.globalPos())

    def update_content(self):
        """
        A new item has been selected, so select the new item.
        """
        prior_selection = self.currentIndex()
        currently_selected = set(obj.row() for obj in self.selectedIndexes())
        ignore_prior_selection = False

        level_objects = self.level_ref.level.get_all_objects()

        labels = [obj.name for obj in level_objects]
        if labels != self.labels:
            self.labels = labels

            self.blockSignals(True)
            self.clear()
            self.addItems(labels)

            for index, level_object in enumerate(level_objects):
                item = self.item(index)
                item.setData(Qt.UserRole, level_object)
                item.setSelected(level_object.selected)
                if level_object.selected and index not in currently_selected:
                    ignore_prior_selection = True

            self.blockSignals(False)

        has_changes = False
        for index, level_object in enumerate(level_objects):
            if level_object.selected and index not in currently_selected:
                has_changes = True
                break
        if not has_changes:
            return

        if self.selectedIndexes():
            self.scrollTo(self.selectedIndexes()[-1])

        if not ignore_prior_selection:
            self.setCurrentIndex(prior_selection)

    def selected_objects(self):
        return [self.item(index.row()).data(Qt.UserRole) for index in self.selectedIndexes()]

    def on_selection_changed(self):
        if self._on_selection_changed_ongoing:
            return

        self._on_selection_changed_ongoing = True
        self.level_ref.selected_objects = self.selected_objects()
        self._on_selection_changed_ongoing = False



from PySide6.QtGui import QPixmap, Qt
from PySide6.QtWidgets import QBoxLayout, QLabel

from foundry import data_dir, get_current_version_name
from foundry.gui.CustomDialog import CustomDialog
from foundry.gui.HorizontalLine import HorizontalLine

LINK_SMB3F = "https://github.com/mchlnix/SMB3-Foundry"
LINK_HUKKA = "http://hukka.ncn.fi/index.php?about"
LINK_SMB3WS = "https://www.romhacking.net/utilities/298/"
LINK_SOUTHBIRD = "https://github.com/captainsouthbird"
LINK_DISASM = "https://github.com/captainsouthbird/smb3"
LINK_BLUEFINCH = "https://www.twitch.tv/bluefinch3000"
LINK_SKY = "https://www.youtube.com/channel/UCnI_HjFGbyRmfOBWzzxK6LA"
LINK_PIJOKRA = "https://github.com/PiJoKra"


class AboutDialog(CustomDialog):
    def __init__(self, parent):
        super(AboutDialog, self).__init__(parent, title="About SMB3Foundry")

        main_layout = QBoxLayout(QBoxLayout.LeftToRight, self)

        image = QPixmap(str(data_dir.joinpath("foundry.ico"))).scaled(200, 200)

        icon = QLabel(self)
        icon.setPixmap(image)

        main_layout.addWidget(icon)

        text_layout = QBoxLayout(QBoxLayout.TopToBottom)

        text_layout.addWidget(QLabel(f"SMB3 Foundry v{get_current_version_name()}", self))
        text_layout.addWidget(HorizontalLine())
        text_layout.addWidget(LinkLabel(self, f'By <a href="{LINK_SMB3F}">Michael</a>'))
        text_layout.addWidget((QLabel("", self)))
        text_layout.addWidget(QLabel("With thanks to:", self))
        text_layout.addWidget(
            LinkLabel(self, f'<a href="{LINK_HUKKA}">Hukka</a> for <a href="{LINK_SMB3WS}">SMB3 Workshop</a>')
        )
        text_layout.addWidget(
            LinkLabel(
                self,
                f'<a href="{LINK_SOUTHBIRD}">Captain Southbird</a> '
                f'for the <a href="{LINK_DISASM}">SMB3 Disassembly</a>',
            )
        )
        text_layout.addWidget(
            LinkLabel(self, f'<a href="{LINK_PIJOKRA}">PiJoKra</a> for helping to parse the disassembly')
        )
        text_layout.addWidget(
            LinkLabel(
                self,
                f'<a href="{LINK_BLUEFINCH}">BlueFinch</a>, ZacMario and '
                f'<a href="{LINK_SKY}">SKJyannick</a> for testing and sanity checking',
            )
        )
        text_layout.addWidget(QLabel("Spinzig for compiling the enemy incompatibilities.", self))

        main_layout.addLayout(text_layout)


class LinkLabel(QLabel):
    def __init__(self, parent, text):
        super(LinkLabel, self).__init__(parent)

        self.setText(text)
        self.setTextFormat(Qt.RichText)
        self.setTextInteractionFlags(Qt.TextBrowserInteraction)
        self.setOpenExternalLinks(True)



from typing import Union

from PySide6.QtCore import QPoint, QSize
from PySide6.QtGui import QCloseEvent, QPainter, QPaintEvent
from PySide6.QtWidgets import (
    QComboBox,
    QHBoxLayout,
    QLayout,
    QStatusBar,
    QToolBar,
    QVBoxLayout,
    QWidget,
)

from foundry.game.gfx.drawable.Block import Block, get_block
from foundry.game.gfx.GraphicsSet import GRAPHIC_SET_NAMES
from foundry.game.gfx.objects.Jump import Jump
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.gfx.objects.LevelObjectFactory import LevelObjectFactory
from foundry.gui.CustomChildWindow import CustomChildWindow
from foundry.gui.LevelSelector import OBJECT_SET_ITEMS
from foundry.gui.Spinner import Spinner
from foundry.gui.util import clear_layout

ID_SPIN_DOMAIN = 1
ID_SPIN_TYPE = 2
ID_SPIN_LENGTH = 3
ID_OBJECT_SET_DROPDOWN = 4
ID_GFX_SET_DROPDOWN = 5

MAX_DOMAIN = 7
MAX_TYPE = 0xFF
MAX_LENGTH = 0xFF


class ObjectViewer(CustomChildWindow):
    def __init__(self, parent):
        super(ObjectViewer, self).__init__(parent, title="Object Viewer")

        self.spin_domain = Spinner(self, MAX_DOMAIN)
        self.spin_domain.valueChanged.connect(self.on_spin)

        self.spin_type = Spinner(self, MAX_TYPE)
        self.spin_type.valueChanged.connect(self.on_spin)

        self.spin_length = Spinner(self, MAX_LENGTH)
        self.spin_length.setDisabled(True)
        self.spin_length.valueChanged.connect(self.on_spin)

        self.toolbar_ = QToolBar(self)

        self.toolbar_.addWidget(self.spin_domain)
        self.toolbar_.addWidget(self.spin_type)
        self.toolbar_.addWidget(self.spin_length)

        self.object_set_dropdown = QComboBox(self.toolbar_)
        self.object_set_dropdown.addItems(OBJECT_SET_ITEMS[1:])
        self.object_set_dropdown.setCurrentIndex(0)

        self.graphic_set_dropdown = QComboBox(self.toolbar_)
        self.graphic_set_dropdown.addItems(GRAPHIC_SET_NAMES)
        self.graphic_set_dropdown.setCurrentIndex(1)

        self.object_set_dropdown.currentIndexChanged.connect(self.on_object_set)
        self.graphic_set_dropdown.currentIndexChanged.connect(self.on_graphic_set)

        self.toolbar_.addWidget(self.object_set_dropdown)
        self.toolbar_.addWidget(self.graphic_set_dropdown)

        self.addToolBar(self.toolbar_)

        self.drawing_area = ObjectDrawArea(self, 1)

        self.status_bar = QStatusBar(parent=self)
        self.status_bar.showMessage(self.drawing_area.current_object.name)

        self.setStatusBar(self.status_bar)

        self.drawing_area.update()

        self.block_list = BlockArray(self, self.drawing_area.current_object)

        central_widget = QWidget()
        central_widget.setLayout(QVBoxLayout())
        central_widget.layout().addWidget(self.drawing_area)
        central_widget.layout().addWidget(self.block_list)

        self.setCentralWidget(central_widget)

        self.layout().setSizeConstraint(QLayout.SetFixedSize)

        return

    def closeEvent(self, event: QCloseEvent):
        self.toolbar_.close()
        super().closeEvent(event)

    def set_object_and_graphic_set(self, object_set: int, graphics_set: int):
        self.object_set_dropdown.setCurrentIndex(object_set - 1)
        self.graphic_set_dropdown.setCurrentIndex(graphics_set)

        self.drawing_area.change_object_set(object_set)
        self.drawing_area.change_graphic_set(graphics_set)

        self.block_list.update_object(self.drawing_area.current_object)
        self.status_bar.showMessage(self.drawing_area.current_object.name)

    def on_object_set(self):
        object_set = self.object_set_dropdown.currentIndex() + 1
        graphics_set = object_set

        self.set_object_and_graphic_set(object_set, graphics_set)

    def on_graphic_set(self):
        object_set = self.object_set_dropdown.currentIndex() + 1
        graphics_set = self.graphic_set_dropdown.currentIndex()

        self.set_object_and_graphic_set(object_set, graphics_set)

    def set_object(self, domain: int, obj_index: int, secondary_length: int):
        object_data = bytearray(4)

        object_data[0] = domain << 5
        object_data[1] = 0
        object_data[2] = obj_index
        object_data[3] = secondary_length

        self.spin_domain.setValue(domain)
        self.spin_type.setValue(obj_index)
        self.spin_length.setValue(secondary_length)

        self.drawing_area.update_object(object_data)
        self.block_list.update_object(self.drawing_area.current_object)

        if self.drawing_area.current_object.is_4byte:
            self.spin_length.setEnabled(True)
        else:
            self.spin_length.setValue(0)
            self.spin_length.setEnabled(False)

        self.drawing_area.update()

        self.status_bar.showMessage(self.drawing_area.current_object.name)

    def on_spin(self, _):
        domain = self.spin_domain.value()
        obj_index = self.spin_type.value()
        secondary_length = self.spin_length.value()

        self.set_object(domain, obj_index, secondary_length)


class ObjectDrawArea(QWidget):
    def __init__(self, parent, object_set, graphic_set=1, palette_index=0):
        super(ObjectDrawArea, self).__init__(parent)

        self.object_factory = LevelObjectFactory(object_set, graphic_set, palette_index, [], False, size_minimal=True)

        self.current_object = self.object_factory.from_data(bytearray([0x0, 0x0, 0x0]), 0)

        self.update_object()

        self.resize(QSize())

    def change_object_set(self, object_set: int):
        self.object_factory.set_object_set(object_set)

        self.update_object()

    def change_graphic_set(self, graphic_set: int):
        self.object_factory.set_graphic_set(graphic_set)
        self.update_object()

    def resize(self, size: QSize):
        if isinstance(self.current_object, Jump):
            return

        self.setMinimumSize(
            QSize(self.current_object.rendered_width * Block.WIDTH, self.current_object.rendered_height * Block.HEIGHT)
        )

    def update_object(self, object_data: Union[bytearray, LevelObject, Jump] = None):
        if object_data is None:
            object_data = self.current_object.data
        elif isinstance(object_data, (LevelObject, Jump)):
            object_data = object_data.data

        self.current_object = self.object_factory.from_data(object_data, 0)

        self.resize(QSize())
        self.update()

    def paintEvent(self, event: QPaintEvent):
        if not isinstance(self.current_object, LevelObject):
            return

        painter = QPainter(self)

        painter.translate(
            QPoint(
                -Block.WIDTH * self.current_object.rendered_base_x, -Block.HEIGHT * self.current_object.rendered_base_y
            )
        )

        self.current_object.draw(painter, Block.WIDTH, transparent=True)


class BlockArray(QWidget):
    def __init__(self, parent, level_object: LevelObject):
        super(BlockArray, self).__init__(parent)

        self.setLayout(QHBoxLayout())

        self.setContentsMargins(0, 0, 0, 0)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        self.level_object = level_object

        self.update_object(level_object)

    def update_object(self, level_object: LevelObject):
        self.level_object = level_object

        clear_layout(self.layout())

        for block_index in self.level_object.blocks:
            block = get_block(
                block_index,
                tuple(tuple(c for c in pal) for pal in self.level_object.palette_group),
                self.level_object.graphics_set,
                bytes(self.level_object.tsa_data),
            )
            self.layout().addWidget(BlockArea(block))

        self.update()


class BlockArea(QWidget):
    def __init__(self, block: Block):
        super(BlockArea, self).__init__()

        self.block = block

        self.setContentsMargins(0, 0, 0, 0)
        self.setToolTip(hex(self.block.index))

    def sizeHint(self):
        return QSize(Block.WIDTH, Block.HEIGHT)

    def paintEvent(self, event):
        painter = QPainter(self)

        self.block.draw(painter, 0, 0, Block.WIDTH)



# Credit to https://doc.qt.io/qtforpython/examples/example_widgets_layouts_flowlayout.html for
# figuring whatever this out

from PySide6.QtCore import QMargins, QPoint, QRect, QSize, Qt
from PySide6.QtWidgets import QLayout, QSizePolicy


class FlowLayout(QLayout):
    def __init__(self, parent=None):
        super().__init__(parent)

        if parent is not None:
            self.setContentsMargins(QMargins(0, 0, 0, 0))

        self._item_list = []

    def fromItems(self, items: list):
        self.clear()
        for item in items:
            self.addItem(item)

    def addItem(self, item):
        self._item_list.append(item)

    def removeItem(self, item_index: int):
        del self._item_list[item_index]

    def clear(self):
        self._item_list = []

    def count(self):
        return len(self._item_list)

    def itemAt(self, index):
        if index >= 0 and index < len(self._item_list):
            return self._item_list[index]

        return None

    def takeAt(self, index):
        if index >= 0 and index < len(self._item_list):
            return self._item_list.pop(index)

        return None

    def expandingDirections(self):
        return Qt.Orientations(Qt.Orientation(0))

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        height = self._do_layout(QRect(0, 0, width, 0), True)
        return height

    def setGeometry(self, rect):
        super(FlowLayout, self).setGeometry(rect)
        self._do_layout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QSize()

        for item in self._item_list:
            size = size.expandedTo(item.minimumSize())

        size += QSize(2 * self.contentsMargins().top(), 2 * self.contentsMargins().top())
        return size

    def _do_layout(self, rect, test_only):
        x = rect.x()
        y = rect.y()
        line_height = 0
        spacing = self.spacing()

        for item in self._item_list:
            style = item.widget().style()
            layout_spacing_x = style.layoutSpacing(QSizePolicy.PushButton, QSizePolicy.PushButton, Qt.Horizontal)
            layout_spacing_y = style.layoutSpacing(QSizePolicy.PushButton, QSizePolicy.PushButton, Qt.Vertical)
            space_x = spacing + layout_spacing_x
            space_y = spacing + layout_spacing_y
            next_x = x + item.sizeHint().width() + space_x
            if next_x - space_x > rect.right() and line_height > 0:
                x = rect.x()
                y = y + line_height + space_y
                next_x = x + item.sizeHint().width() + space_x
                line_height = 0

            if not test_only:
                item.setGeometry(QRect(QPoint(x, y), item.sizeHint()))

            x = next_x
            line_height = max(line_height, item.sizeHint().height())

        return y + line_height - rect.y()



from pydantic import BaseModel

from foundry.core.warnings.util import type_to_pydantic_warning
from foundry.core.warnings.Warning import PydanticWarning
from foundry.core.warnings.WarningType import WarningType


class WarningCreator(BaseModel):
    def __init_subclass__(cls, **kwargs) -> None:
        return super().__init_subclass__(**kwargs)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def generate_warning(cls, v: dict) -> PydanticWarning:
        """
        The constructor for each specific warning.

        Parameters
        ----------
        v : dict
            The dictionary to create the warning.

        Returns
        -------
        Warning
            The created warning as defined by `v["type"]`

        Raises
        ------
        NotImplementedError
            If the constructor does not have a valid constructor for `v["type"]`.
        """
        type_ = WarningType(v["type"])
        warnings = type_to_pydantic_warning()
        if type_ not in warnings:
            raise NotImplementedError(f"There is no layout of type {type_}")

        return warnings[type_](**v)

    @classmethod
    def validate(cls, v):
        if not isinstance(v, dict):
            raise TypeError("Dictionary required")
        if "type" not in v:
            raise TypeError("Must have a type")
        if not WarningType.has_value(type_ := v["type"]):
            raise TypeError(f"{type_} is not a valid warning type")
        return cls.generate_warning(v)



from abc import ABC, abstractmethod

from PySide6.QtCore import QRect

from foundry.core.Position import PositionProtocol
from foundry.game.Definitions import Definition

EXPANDS_NOT = 0b00
EXPANDS_HORIZ = 0b01
EXPANDS_VERT = 0b10
EXPANDS_BOTH = EXPANDS_HORIZ | EXPANDS_VERT


class ObjectLike(ABC):
    obj_index: int
    name: str

    rect: QRect

    @abstractmethod
    def render(self):
        pass

    @abstractmethod
    def draw(self, dc, zoom, transparent):
        pass

    @abstractmethod
    def get_status_info(self):
        pass

    @abstractmethod
    def move_by(self, dx, dy):
        pass

    @property
    @abstractmethod
    def definition(self) -> Definition:
        pass

    @property
    @abstractmethod
    def position(self) -> PositionProtocol:
        pass

    @position.setter
    @abstractmethod
    def position(self, position: PositionProtocol) -> None:
        pass

    @abstractmethod
    def resize_by(self, dx, dy):
        pass

    @abstractmethod
    def point_in(self, x, y):
        pass

    def get_rect(self, block_length=1) -> QRect:
        if block_length != 1:
            x, y = self.rect.topLeft().toTuple()
            w, h = self.rect.size().toTuple()

            x *= block_length
            w *= block_length
            y *= block_length
            h *= block_length

            return QRect(x, y, w, h)
        else:
            return self.rect

    @abstractmethod
    def change_type(self, new_type):
        pass

    @abstractmethod
    def __contains__(self, point):
        pass

    @abstractmethod
    def to_bytes(self):
        pass

    def expands(self):
        return EXPANDS_NOT

    def primary_expansion(self):
        return EXPANDS_NOT



from typing import Optional

from attr import attrs
from PySide6.QtCore import Signal, SignalInstance
from PySide6.QtGui import QCloseEvent, Qt
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QFormLayout,
    QGridLayout,
    QLayout,
    QStatusBar,
    QToolBar,
    QWidget,
)

from foundry import icon
from foundry.core.player_animations import ANIMATION_WIDTH
from foundry.core.player_animations.PlayerAnimation import PlayerAnimation
from foundry.core.player_animations.util import (
    get_animations_palette_index,
    load_player_animation_data,
    load_player_animations,
    load_power_up_offsets,
    load_power_up_palettes,
    save_player_animations_to_rom,
)
from foundry.core.Position import Position
from foundry.core.sprites import SPRITE_SIZE
from foundry.core.sprites.Sprite import Sprite, SpriteProtocol
from foundry.core.sprites.SpriteGroup import SpriteGroup, SpriteGroupProtocol
from foundry.core.UndoController import UndoController
from foundry.game.gfx.Palette import PaletteGroupProtocol, PaletteProtocol
from foundry.gui.CustomChildWindow import CustomChildWindow
from foundry.gui.PaletteEditorWidget import PaletteEditorWidget
from foundry.gui.PlayerFrameEditor import (
    PlayerFrameEditorController as PlayerFrameEditor,
)
from foundry.gui.SettingsDialog import SettingsDialog
from foundry.gui.Spinner import Spinner
from foundry.gui.SpriteViewerWidget import SpriteViewerWidget

POWERUPS = [
    ("Small Mario", 32, 53),
    ("Big Mario", 6, 48),
    ("Fire Mario", 16, 53),
    ("Raccoon Mario", 57, 53),
    ("Frog Mario", 56, 53),
    ("Tanooki Mario", 54, 53),
    ("Hammer Mario", 58, 53),
]


@attrs(slots=True, auto_attribs=True)
class PlayerViewerModel:
    is_mario: bool
    power_up: int
    power_up_offsets: list[int]
    palette_group: PaletteGroupProtocol
    animations: list[PlayerAnimation]

    @classmethod
    def from_rom(
        cls,
        is_mario: bool = True,
        power_up: int = 0,
        power_up_offsets: Optional[list[int]] = None,
        palette_group: Optional[PaletteGroupProtocol] = None,
        animations: Optional[list[PlayerAnimation]] = None,
    ):
        return cls(
            is_mario,
            power_up,
            power_up_offsets or load_power_up_offsets(),
            palette_group or load_power_up_palettes(),
            animations or load_player_animations(),
        )

    def to_bytes(self) -> tuple[bytes, bytes, bytes, bytes]:
        animation_bytes = bytearray()
        page_offset_bytes = bytearray()
        for animation in self.animations:
            ani_bytes, page_bytes = animation.to_bytes()
            ani_bytes = bytearray([i + 1 for i in ani_bytes])
            animation_bytes += ani_bytes
            page_offset_bytes += page_bytes

        return (
            bytes(o & 0xFF for o in self.power_up_offsets),
            bytes(self.palette_group),
            bytes(animation_bytes),
            bytes(page_offset_bytes),
        )


class PlayerViewerController(CustomChildWindow):
    power_up_changed: SignalInstance = Signal(int)  # type: ignore
    is_mario_changed: SignalInstance = Signal(bool)  # type: ignore
    palette_group_changed: SignalInstance = Signal(PaletteGroupProtocol)  # type: ignore
    power_up_offsets_changed: SignalInstance = Signal(object)  # type: ignore
    destroyed: SignalInstance = Signal()  # type: ignore

    def __init__(
        self,
        parent: Optional[QWidget],
        title: str = "Player Viewer",
        is_mario: bool = True,
        power_up: int = 0,
        power_up_offsets: Optional[list[int]] = None,
        palette_group: Optional[PaletteGroupProtocol] = None,
        animations: Optional[list[PlayerAnimation]] = None,
    ):
        super().__init__(parent, title)

        self.model = PlayerViewerModel.from_rom(is_mario, power_up, power_up_offsets, palette_group, animations)
        self.view = PlayerViewerView(self, self.sprite_groups)
        self.undo_controller = UndoController(self.model.to_bytes())
        self.frame_editor: Optional[PlayerFrameEditor] = None
        self.frame_editor_index = 0
        self.setCentralWidget(self.view)
        self.toolbar = QToolBar(self)
        self.side_toolbar = QToolBar(self)

        self.toolbar.setMovable(False)
        self.side_toolbar.setMovable(False)

        self.view.mouse_moved_over_widget.connect(self._on_mouse_move)
        self.view.clicked.connect(self._show_player_frame_editor)

        self.undo_action = self.toolbar.addAction(icon("rotate-ccw.svg"), "Undo Action")
        self.redo_action = self.toolbar.addAction(icon("rotate-cw.svg"), "Redo Action")
        self.zoom_out_action = self.toolbar.addAction(icon("zoom-out.svg"), "Zoom Out")
        self.zoom_in_action = self.toolbar.addAction(icon("zoom-in.svg"), "Zoom In")

        self.powerup_combo_box = QComboBox()
        for (name, x, y) in POWERUPS:
            powerup_icon = SettingsDialog._load_from_png(x, y)
            self.powerup_combo_box.addItem(powerup_icon, name)
        self.is_mario_check_box = QCheckBox()
        self.palette_editor = PaletteEditorWidget(self, self.palette)
        self.power_up_offset_spinner = Spinner(self, 0xFF)

        def change_zoom(offset: int):
            self.view.zoom += offset

        def undo(*_):
            self.undo_controller.undo()
            self._update()

        def redo(*_):
            self.undo_controller.redo()
            self._update()

        def set_power_up(value: int):
            self.power_up = value

        def set_is_mario(value: bool):
            self.is_mario = value

        def set_palette(value: PaletteProtocol):
            self.palette = value

        def set_power_up_offset(value: int):
            self.power_up_offset = value

        self.undo_action.triggered.connect(undo)  # type: ignore
        self.undo_action.setEnabled(False)
        self.redo_action.triggered.connect(redo)  # type: ignore
        self.redo_action.setEnabled(False)
        self.zoom_out_action.triggered.connect(lambda *_: change_zoom(-1))  # type: ignore
        self.zoom_in_action.triggered.connect(lambda *_: change_zoom(1))  # type: ignore

        self.powerup_combo_box.currentIndexChanged.connect(set_power_up)  # type: ignore
        self.powerup_combo_box.setCurrentIndex(power_up)

        self.is_mario_check_box.setChecked(is_mario)
        self.is_mario_check_box.toggled.connect(set_is_mario)  # type: ignore

        self.palette_editor.palette_changed.connect(set_palette)

        self.power_up_offset_spinner.valueChanged.connect(set_power_up_offset)  # type: ignore

        side_toolbar_layout = QFormLayout(self)
        side_toolbar_layout.addRow("Power Up", self.powerup_combo_box)
        side_toolbar_layout.addRow("Mario or Luigi", self.is_mario_check_box)
        side_toolbar_layout.addRow("Palette Editor", self.palette_editor)
        side_toolbar_layout.addRow("Page Offset", self.power_up_offset_spinner)

        class LayoutWidget(QWidget):
            def __init__(self, parent, layout: QLayout):
                super().__init__(parent)
                self.setLayout(layout)

        self.side_toolbar.addWidget(LayoutWidget(self.side_toolbar, side_toolbar_layout))

        self.addToolBar(self.toolbar)
        self.addToolBar(Qt.RightToolBarArea, self.side_toolbar)

        self.setStatusBar(QStatusBar(self))
        self.layout().setSizeConstraint(QLayout.SetFixedSize)

    def closeEvent(self, event: QCloseEvent):
        if self.frame_editor is not None:
            self.frame_editor.close()
        self.destroyed.emit()
        super().closeEvent(event)

    @property
    def power_up(self) -> int:
        return self.model.power_up

    @power_up.setter
    def power_up(self, power_up: int):
        self.model.power_up = power_up
        self._update_palette_editor()
        self._update_view_animations()
        self.power_up_offset_spinner.setValue(self.power_up_offset)
        self.powerup_combo_box.setCurrentIndex(power_up)
        self.power_up_changed.emit(power_up)

    @property
    def is_mario(self) -> bool:
        return self.model.is_mario

    @is_mario.setter
    def is_mario(self, is_mario: bool):
        self.model.is_mario = is_mario
        self._update()
        self.is_mario_check_box.setChecked(is_mario)
        self.is_mario_changed.emit(is_mario)

    @property
    def power_up_offset(self) -> int:
        return self.power_up_offsets[self.power_up]

    @power_up_offset.setter
    def power_up_offset(self, offset: int):
        offsets = self.power_up_offsets
        offsets[self.power_up] = offset
        self.power_up_offsets = offsets

    @property
    def power_up_offsets(self) -> list[int]:
        return self.model.power_up_offsets

    @power_up_offsets.setter
    def power_up_offsets(self, offsets: list[int]):
        self.model.power_up_offsets = offsets
        self.undo_controller.do(self.model.to_bytes())
        self.update()
        self.power_up_offsets_changed.emit(offsets)

    @property
    def palette(self) -> PaletteProtocol:
        return self.palette_group[get_animations_palette_index(self.is_mario, self.power_up)]

    @palette.setter
    def palette(self, palette: PaletteProtocol):
        palette_group = self.palette_group
        palette_group[get_animations_palette_index(self.is_mario, self.power_up)] = palette
        self.palette_group = palette_group

    @property
    def palette_group(self) -> PaletteGroupProtocol:
        return self.model.palette_group

    @palette_group.setter
    def palette_group(self, palette_group: PaletteGroupProtocol):
        self.model.palette_group = palette_group
        self.undo_controller.do(self.model.to_bytes())
        self._update()
        self.palette_group_changed.emit(palette_group)

    @property
    def sprite_groups(self) -> list[SpriteGroupProtocol]:
        sprite_groups: list[SpriteGroupProtocol] = []

        for animation in load_player_animation_data(
            self.model.animations,
            self.model.palette_group,
            self.model.is_mario,
            self.model.power_up,
            self.model.power_up_offsets,
        ):
            sprites: list[SpriteProtocol] = []

            for idx, sprite in enumerate(animation.frames):
                sprites.append(
                    Sprite(
                        Position(
                            (idx % ANIMATION_WIDTH) * SPRITE_SIZE.width,
                            (idx // ANIMATION_WIDTH) * SPRITE_SIZE.height,
                        ),
                        sprite,
                        animation.palette_index,
                        animation.horizontal_flip[idx],
                        do_not_render=sprite == 0xF0,
                    )
                )

            sprite_groups.append(SpriteGroup(Position(0, 0), sprites, animation.graphics_set, animation.palette_group))

        return sprite_groups

    def _update(self):
        self.undo_action.setEnabled(self.undo_controller.can_undo)
        self.redo_action.setEnabled(self.undo_controller.can_redo)
        save_player_animations_to_rom(*self.model.to_bytes())
        self._update_palette_editor()
        self._update_view_animations()
        if self.frame_editor is not None:
            self.frame_editor.is_mario = self.is_mario

    def _update_palette_editor(self):
        self.palette_editor._palette = self.palette
        self.palette_editor._update()

    def _update_view_animations(self):
        self.view.sprite_groups = self.sprite_groups
        self.view._update()

    def _show_player_frame_editor(self, index: int):
        self.frame_editor_index = index

        if self.frame_editor is None:
            self.frame_editor = PlayerFrameEditor(
                self, self.model.animations[index], self.power_up_offsets, self.palette_group, self.is_mario
            )

            def remove_frame_editor():
                self.frame_editor = None

            def change_frame_offset(offset: int):
                self.model.animations[self.frame_editor_index].offset = offset
                self.undo_controller.do(self.model.to_bytes())
                self._update()

            def change_frame_sprites(sprites: bytearray):
                self.model.animations[self.frame_editor_index].animations = sprites
                self.undo_controller.do(self.model.to_bytes())
                self._update()

            self.frame_editor.destroyed.connect(remove_frame_editor)
            self.frame_editor.offset_changed.connect(change_frame_offset)
            self.frame_editor.frames_changed.connect(change_frame_sprites)
            self.frame_editor.show()
        else:
            self.frame_editor.model.animation = self.model.animations[index]
            self.frame_editor._update_view_animations()

    def _on_mouse_move(self, x: int, y: int, index: int, hex_index: str):
        self.statusBar().showMessage(f"Row: {y}, Column: {x}, Index: {index} / {hex_index}")


class PlayerViewerView(QWidget):
    mouse_moved_over_widget: SignalInstance = Signal(int, int, int, str)  # type: ignore
    clicked: SignalInstance = Signal(int)  # type: ignore

    ANIMATIONS_PER_ROW = 9
    ANIMATIONS_PER_COLUMN = 9

    def __init__(
        self,
        parent: Optional[QWidget],
        sprite_groups: list[SpriteGroupProtocol],
        zoom: int = 2,
    ):
        super().__init__(parent)

        self.sprite_groups = sprite_groups
        self._zoom = zoom

        self.layout_ = QGridLayout(self)
        self.widgets: list[SpriteViewerWidget] = []

        for idx, sprite_group in enumerate(sprite_groups):
            widget = SpriteViewerWidget(self, sprite_group, zoom=self.zoom)
            widget.mouse_moved_over_widget.connect(
                lambda *_, idx=idx: self.mouse_moved_over_widget.emit(
                    idx % self.ANIMATIONS_PER_ROW,
                    idx // self.ANIMATIONS_PER_ROW,
                    idx,
                    hex(idx).upper().replace("X", "x"),
                )
            )
            widget.clicked.connect(lambda *_, idx=idx: self.clicked.emit(idx))
            self.widgets.append(widget)
            self.layout_.addWidget(widget, idx // self.ANIMATIONS_PER_ROW, idx % self.ANIMATIONS_PER_ROW)

        self.setLayout(self.layout_)

    @property
    def zoom(self) -> int:
        return self._zoom

    @zoom.setter
    def zoom(self, value: int):
        self._zoom = value
        for widget in self.widgets:
            widget.zoom = value

    def _update(self):
        for (widget, sprite_group) in zip(self.widgets, self.sprite_groups):
            widget.sprite_group = sprite_group
            widget.update()



from attr import attrs


@attrs(slots=True, auto_attribs=True)
class PlayerAnimation:
    """
    A representation of a single player animation.

    Attributes
    ----------
    animations: bytearray
        The frames inside animation.
    offset: int
        The page offset of its graphics, which applies to each power up of the player's.
    """

    animations: bytearray
    offset: int

    def to_bytes(self) -> tuple[bytes, bytes]:
        return self.animations, bytes([self.offset & 0xFF])



from PySide6.QtCore import QRect
from PySide6.QtGui import QImage

from foundry import data_dir
from foundry.game.gfx.drawable.Block import Block
from foundry.game.gfx.objects.EnemyItem import EnemyObject
from foundry.game.gfx.Palette import PALETTE_GROUPS_PER_OBJECT_SET, PaletteGroup


class EnemyItemFactory:
    object_set: int
    graphic_set: int

    definitions: list = []

    def __init__(self, object_set: int, palette_index: int):
        png = QImage(str(data_dir.joinpath("gfx.png")))

        png.convertTo(QImage.Format_RGB888)

        rows_per_object_set = 256 // 64

        y_offset = 12 * rows_per_object_set * Block.HEIGHT

        self.png_data = png.copy(QRect(0, y_offset, png.width(), png.height() - y_offset))

        self.palette_group = PaletteGroup.from_tileset(object_set, PALETTE_GROUPS_PER_OBJECT_SET + palette_index)

    def from_data(self, data, _):
        return EnemyObject(data, self.png_data, self.palette_group)

    def from_properties(self, enemy_item_id: int, x: int, y: int):
        data = bytearray(3)

        data[0] = enemy_item_id
        data[1] = x
        data[2] = y

        obj = self.from_data(data, 0)

        return obj



from typing import Optional

from PySide6.QtCore import QSize, Qt, Signal, SignalInstance
from PySide6.QtGui import QColor, QMouseEvent, QPainter, QPaintEvent
from PySide6.QtWidgets import QFrame, QSizePolicy, QWidget

from foundry.core.sprites import SPRITE_SIZE
from foundry.core.sprites.Sprite import SpriteProtocol
from foundry.core.sprites.SpriteGroup import SpriteGroupProtocol
from foundry.game.gfx.drawable import MASK_COLOR
from foundry.game.gfx.GraphicsSet import GraphicsSetProtocol
from foundry.game.gfx.Palette import PaletteGroupProtocol


class SpriteViewerWidget(QFrame):
    sprites_changed: SignalInstance = Signal(list[SpriteProtocol])  # type: ignore
    graphics_set_changed: SignalInstance = Signal(GraphicsSetProtocol)  # type: ignore
    palette_group_changed: SignalInstance = Signal(PaletteGroupProtocol)  # type: ignore
    transparency_changed: SignalInstance = Signal(bool)  # type: ignore
    zoom_changed: SignalInstance = Signal(int)  # type: ignore
    mouse_moved_over_widget: SignalInstance = Signal(QMouseEvent)  # type: ignore
    clicked: SignalInstance = Signal(int, int)  # type: ignore

    def __init__(
        self, parent: Optional[QWidget], sprite_group: SpriteGroupProtocol, zoom: int = 2, transparent: bool = True
    ):
        super().__init__(parent)

        self.setMouseTracking(True)

        self.sprite_group = sprite_group
        self._transparent = transparent
        self.zoom = zoom

        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

    @property
    def sprites(self) -> list[SpriteProtocol]:
        return self.sprite_group.sprites

    @sprites.setter
    def sprites(self, sprites: list[SpriteProtocol]):
        self.sprite_group.sprites = sprites
        self.sprites_changed.emit(sprites)
        self.update()

    @property
    def graphics_set(self) -> GraphicsSetProtocol:
        return self.sprite_group.graphics_set

    @graphics_set.setter
    def graphics_set(self, graphics_set: GraphicsSetProtocol):
        self.sprite_group.graphics_set = graphics_set
        self.graphics_set_changed.emit(graphics_set)
        self.update()

    @property
    def palette_group(self) -> PaletteGroupProtocol:
        return self.sprite_group.palette_group

    @palette_group.setter
    def palette_group(self, palette_group: PaletteGroupProtocol):
        self.sprite_group.palette_group = palette_group
        self.palette_group_changed.emit(palette_group)
        self.update()

    @property
    def transparent(self) -> bool:
        return self._transparent

    @transparent.setter
    def transparent(self, transparent: bool):
        self._transparent = transparent
        self.transparency_changed.emit(transparent)
        self.update()

    @property
    def zoom(self) -> int:
        return self._zoom

    @zoom.setter
    def zoom(self, zoom: int):
        self._zoom = zoom
        self.setFixedSize(self.sizeHint())
        self.zoom_changed.emit(zoom)
        self.update()

    def sizeHint(self):
        return QSize(self.sprite_group.size.width * self.zoom, self.sprite_group.size.height * self.zoom)

    def mouseMoveEvent(self, event: QMouseEvent):
        self.mouse_moved_over_widget.emit(event)

    def mouseReleaseEvent(self, event: QMouseEvent):
        x, y = event.x() // (self.zoom * SPRITE_SIZE.width), event.y() // (self.zoom * SPRITE_SIZE.height)
        self.clicked.emit(x, y)

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)

        image = self.sprite_group.image(self.zoom)
        if self.transparent:
            mask = image.createMaskFromColor(QColor(*MASK_COLOR).rgb(), Qt.MaskOutColor)
            image.setAlphaChannel(mask)

        painter.drawImage(0, 0, image)

        return super().paintEvent(event)



from foundry.core.warnings.Warning import PydanticWarning, Warning
from foundry.game.gfx.objects.ObjectLike import ObjectLike


class OutsideLevelBoundsWarning(Warning):
    """
    A warning for objects going outside the level bounds.
    """

    def check_object(self, obj: ObjectLike, level=None, *args, **kwargs) -> bool:
        """
        Determines if an object should emit a warning for being outside the level bounds.

        Parameters
        ----------
        obj : ObjectLike
            The object to check.
        level : Optional[Level]
            The level to check the bounds of.

        Returns
        -------
        bool
            If the object should emit a warning.

        Notes
        -----
        If not level is provided, no warning will be emitted.
        """
        if level is None:
            return False
        return not level.get_rect().contains(obj.get_rect())

    def get_message(self, obj: ObjectLike) -> str:
        return f"{obj} is outside of level bounds."


class PydanticOutsideLevelBoundsWarning(PydanticWarning):
    """
    A JSON model of a warning that ensures that an object is inside the bounds of the level.
    """



"""
A series of constants useful for acquiring the player's animations.
"""

POWER_UP_OFFSETS = [0x50, 0x54, 0x54, 0x00, 0x50, 0x40, 0x44]
PLAYER_FRAME_START = 0x3AC61
PLAYER_FRAME_PAGE_OFFSET = 0x3AE47
PLAYER_SUIT_PAGE_OFFSET = 0x3AEA8
PLAYER_FRAMES = 81
PLAYER_POWER_UPS = 7
SPRITES_PER_FRAME = 6
PLAYER_POWER_UPS_PALETTE_COUNT = 8
PLAYER_POWER_UPS_PALETTES = 0x10539
ANIMATION_WIDTH = 3
ANIMATION_HEIGHT = 2



from typing import Protocol

from attr import attrs


class DataProtocol(Protocol):
    """
    A representation of bytes inside a given file.

    Attributes
    ----------
    location: int
        The starting point of the data.
    data: bytes
        The data at that location.
    """

    location: int
    data: bytes


@attrs(slots=True, auto_attribs=True)
class Data:
    """
    A representation of bytes inside a given file, that uses ``attrs`` to create a basic implementation.

    Attributes
    ----------
    location: int
        The starting point of the data.
    data: bytes
        The data at that location.
    """

    location: int
    data: bytes



from PySide6.QtCore import QPoint, QSize

from foundry.game.File import ROM
from foundry.game.gfx.drawable.Block import Block
from foundry.game.gfx.GraphicsSet import GraphicsSet
from foundry.game.gfx.objects.MapObject import MapObject
from foundry.game.gfx.Palette import PaletteGroup
from foundry.game.level.LevelLike import LevelLike
from foundry.smb3parse.levels.world_map import (
    WORLD_MAP_HEIGHT,
    WORLD_MAP_SCREEN_SIZE,
    WORLD_MAP_SCREEN_WIDTH,
)
from foundry.smb3parse.levels.world_map import WorldMap as _WorldMap
from foundry.smb3parse.objects.object_set import WORLD_MAP_OBJECT_SET

OVERWORLD_GRAPHIC_SET = 0


class WorldMap(LevelLike):
    def __init__(self, world_index):
        self._internal_world_map = _WorldMap.from_world_number(ROM(), world_index)

        super(WorldMap, self).__init__(0, self._internal_world_map.layout_address)

        self.name = f"World {world_index} - Overworld"

        self.graphics_set = GraphicsSet.from_tileset(OVERWORLD_GRAPHIC_SET)
        self.palette_group = PaletteGroup.from_tileset(WORLD_MAP_OBJECT_SET, 0)

        self.object_set = WORLD_MAP_OBJECT_SET
        self.tsa_data = ROM.get_tsa_data(self.object_set)

        self.world = 0
        self.level_number = world_index

        self.objects = []

        self._load_objects()

        self._calc_size()

    def _load_objects(self):
        self.objects.clear()

        for index, world_position in enumerate(self._internal_world_map.gen_positions()):
            screen_offset = (index // WORLD_MAP_SCREEN_SIZE) * WORLD_MAP_SCREEN_WIDTH

            x = screen_offset + (index % WORLD_MAP_SCREEN_WIDTH)
            y = (index // WORLD_MAP_SCREEN_WIDTH) % WORLD_MAP_HEIGHT

            block = Block(
                world_position.tile(),
                tuple(tuple(c for c in pal) for pal in self.palette_group),
                self.graphics_set,
                self.tsa_data,
            )

            self.objects.append(MapObject(block, x, y))

        assert len(self.objects) % WORLD_MAP_HEIGHT == 0

    def _calc_size(self):
        self.width = len(self.objects) // WORLD_MAP_HEIGHT
        self.height = WORLD_MAP_HEIGHT

        self.size = self.width, self.height

    def add_object(self, obj, _):
        self.objects.append(obj)

        self.objects.sort(key=self._array_index)

    @property
    def q_size(self):
        return QSize(*self.size) * Block.SIDE_LENGTH

    @staticmethod
    def _array_index(obj):
        return obj.y_position * WORLD_MAP_SCREEN_WIDTH + obj.x_position

    def get_object_names(self):
        return [obj.name for obj in self.objects]

    def draw(self, dc, zoom, transparency=None, show_expansion=None):
        for obj in self.objects:
            obj.draw(dc, Block.SIDE_LENGTH * zoom, transparency)

    def index_of(self, obj):
        return self.objects.index(obj)

    def get_all_objects(self):
        return self.objects

    def object_at(self, x, y):
        point = QPoint(x, y)

        for obj in reversed(self.objects):
            if obj.rect.contains(point):
                return obj

        return None

    def to_bytes(self):
        return_array = bytearray(len(self.objects))

        for obj in self.objects:
            index = self._array_index(obj)

            return_array[index] = obj.to_bytes()

        return self.layout_address, return_array

    def from_bytes(self, data, _=None):
        offset, obj_bytes = data

        self.layout_address = offset
        self._load_objects(obj_bytes)

        self._calc_size()

    def get_object(self, index):
        return self.objects[index]

    def remove_object(self, obj):
        self.objects.remove(obj)

    def level_at_position(self, x: int, y: int):
        assert isinstance(x, int)
        assert isinstance(y, int)

        screen = x // WORLD_MAP_SCREEN_WIDTH + 1

        x %= WORLD_MAP_SCREEN_WIDTH

        return self._internal_world_map.level_for_position(screen, y, x)

    def level_name_at_position(self, x: int, y: int) -> str:
        screen = x // WORLD_MAP_SCREEN_WIDTH + 1

        x %= WORLD_MAP_SCREEN_WIDTH

        return self._internal_world_map.level_name_for_position(screen, y, x)



from collections import namedtuple
from typing import NamedTuple


def read_string_dict(path):
    return_dict = dict()

    with open(path, "r") as f:
        for line in f.readlines():
            key, value = line.rstrip().split("=")

            return_dict[key] = value

    return return_dict


MAX_LEVEL_SECTIONS = 65
SMB3_LEVEL_COUNT = 298

MapscreenPointerLocation = namedtuple("MapscreenPointerLocation", "count offset")
ObjectInfo = namedtuple("ObjectInfo", "index subindex x y width height x2 y2 obj objtype rect drag")


class Mario3Level(NamedTuple):
    game_world: int
    level_in_world: int
    rom_level_offset: int
    enemy_offset: int
    real_obj_set: int
    name: str


map_sprite_names = [
    "Nothing?",
    '"Help!"',
    "Ship",
    "Hammer Bros.",
    "Boomerang Bros.",
    "Sledge Bros.",
    "Fire Bros.",
    "Piranha Plant",
    "Weird",
    "N-card",
    "White Mushroom House",
    "Coin Ship",
    "World 8 Ship #1",
    "Battleship",
    "Tank",
    "World 8 Ship #2",
    "Boat",
]

obj_sets = [
    "Map Screen",
    "Plains Level",
    "Hilly/Underground Level",
    "Sky Level",
    "Dungeon",
    "Airship",
    "Cloudy Level",
    "Desert Level",
    "Water/Pipe Level",
    "Giant Level",
    "Ice Level",
]

mushroom_houses = [
    "P-Wing Only",
    "Warp Whistle Only",
    "P-Wing Only",
    "Frog Suit Only",
    "Tanooki Suit Only",
    "Hammer Suit Only",
    "Frog, Tanooki, Hammer Suit",
    "Mushroom, Leaf, Flower",
    "Leaf, Flower, Frog Suit",
    "Leaf, Flower, Tanooki Suit",
    "Anchor Only",
    "Warp Whistle, P-Wing, Frog Suit",
    "Frog Suit, P-Wing, Tanooki Suit",
    "Frog, Tanooki, Hammer Suit",
    "Warp Whistle, P-Wing, Frog Suit",
    "Frog Suit, P-Wing, Tanooki Suit",
    "Frog, Tanooki, Hammer Suit",
    "Warp Whistle, P-Wing, Frog Suit",
    "Frog Suit, P-Wing, Tanooki Suit",
    "Frog, Tanooki, Hammer Suit",
    "Warp Whistle, P-Wing, Frog Suit",
    "Frog Suit, P-Wing, Tanooki Suit",
    "Frog, Tanooki, Hammer Suit",
    "Warp Whistle, P-Wing, Frog Suit",
    "Frog Suit, P-Wing, Tanooki Suit",
    "Frog, Tanooki, Hammer Suit",
    "Warp Whistle, P-Wing, Frog Suit",
]

map_pointers = [
    MapscreenPointerLocation(21, 0x19438),  # World 1
    MapscreenPointerLocation(47, 0x194BA),  # World 2
    MapscreenPointerLocation(52, 0x195D8),  # World 3
    MapscreenPointerLocation(34, 0x19714),  # World 4
    MapscreenPointerLocation(42, 0x197E4),  # World 5
    MapscreenPointerLocation(57, 0x198E4),  # World 6
    MapscreenPointerLocation(46, 0x19A3E),  # World 7
    MapscreenPointerLocation(41, 0x19B56),  # World 8
    MapscreenPointerLocation(10, 0x19C50),  # Warp Zone (no enemies or level pointers)
]



from PySide6.QtCore import Signal, SignalInstance
from PySide6.QtGui import QContextMenuEvent
from PySide6.QtWidgets import QListWidget, QMenu, QWidget

from foundry.game.level.LevelRef import LevelRef

ID_ADD_JUMP = 1
ID_DEL_JUMP = 2
ID_EDIT_JUMP = 3


class JumpList(QListWidget):
    add_jump: SignalInstance = Signal()
    edit_jump: SignalInstance = Signal()
    remove_jump: SignalInstance = Signal()

    def __init__(self, parent: QWidget, level_ref: LevelRef):
        super(JumpList, self).__init__(parent)

        self._level_ref = level_ref

        self._level_ref.data_changed.connect(self.update)
        self.itemDoubleClicked.connect(lambda _: self.edit_jump.emit())
        self.add_jump.connect(self.scrollToBottom)

        self.setWhatsThis(
            "<b>Jump List</b><br/>"
            "Every level can designate another level to jump to, in case a pipe or a door is entered. This is done in "
            "the header, which can be edited with the Header Editor. While only one such level can be defined, where "
            "and how to enter that level can be defined multiple times with multiple jumps.<br/>"
            "A jump is valid for one screen, a 16-block wide/high section of the level, depending on if the level is "
            "vertical or not, and all objects within that section, capable of handling a jump, will jump to the same "
            "position in the same way. To see where these jump zones are, enable the Jump Zone option in the View menu."
            "<br/><br/>"
            "Tip: By having multiple jumps with different entry positions, you could make it look, like you are "
            "jumping to two different levels, when, in fact, you are jumping to two different sections of the same "
            "level."
        )

    def update(self):
        self.clear()

        jumps = self._level_ref.level.jumps

        self.addItems([str(jump) for jump in jumps])

    def delete(self):
        currently_selected = set(obj.row() for obj in self.selectedIndexes())
        jumps = self._level_ref.level.jumps
        for idx, jump in enumerate(jumps):
            if idx in currently_selected:
                self._level_ref.level.remove_jump(jump)

    def contextMenuEvent(self, event: QContextMenuEvent):
        item = self.itemAt(event.pos())

        menu = QMenu()

        if item is None:
            add_action = menu.addAction("Add Jump")
            add_action.triggered.connect(self.add_jump.emit)

        else:
            edit_action = menu.addAction("Edit Jump")
            edit_action.triggered.connect(self.edit_jump.emit)

            remove_action = menu.addAction("Remove Jump")
            remove_action.triggered.connect(self.remove_jump.emit)

        menu.exec_(event.globalPos())



from typing import Optional

from PySide6.QtWidgets import QDialog, QWidget

from foundry.game.gfx.Palette import PaletteProtocol
from foundry.gui.ColorSelector import ColorSelector
from foundry.gui.PaletteWidget import PaletteWidget


class PaletteEditorWidget(PaletteWidget):
    def __init__(self, parent: Optional[QWidget], palette: PaletteProtocol):
        super().__init__(parent, palette)

        for idx, btn in enumerate(self._buttons):
            btn.clicked.connect(lambda idx=idx: self._open_color_selector(idx))

    def _open_color_selector(self, button_index: int):
        selector = ColorSelector(self)

        if QDialog.Accepted == selector.exec_():
            palette = self.palette
            palette[button_index] = selector.last_selected_color_index
            self.palette = palette



from foundry.core.warnings.Warning import PydanticWarning, Warning
from foundry.game.gfx.objects.Jump import Jump
from foundry.game.gfx.objects.ObjectLike import ObjectLike


class InvalidWarpWarning(Warning):
    """
    A warning for jumps not having a valid warp.
    """

    def check_object(self, obj: Jump, level=None, *args, **kwargs) -> bool:
        """
        Determines if a jump should emit a warning for not having a place to warp to.

        Parameters
        ----------
        obj : Jump
            The jump to check.
        level : Optional[Level]
            The level to check the warps of.

        Returns
        -------
        bool
            If the jump should emit a warning.

        Notes
        -----
        If not level is provided, no warning will be emitted.
        """
        return level is not None and not level.has_next_area

    def get_message(self, obj: ObjectLike) -> str:
        return f"Level has {obj}, but no Jump Destination in Level Header."


class PydanticInvalidWarpWarning(PydanticWarning):
    """
    A JSON model of a warning that ensures that for a given jump it has a place to warp to.
    """



from typing import Union

from PySide6.QtCore import Signal, SignalInstance
from PySide6.QtWidgets import QScrollArea, QTabWidget

from foundry.game.gfx.objects.EnemyItem import EnemyObject
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.gui.ObjectIcon import ObjectButton
from foundry.gui.ObjectToolBox import ObjectToolBox


class TabbedToolBox(QTabWidget):
    selected: SignalInstance = Signal(ObjectButton)  # type: ignore
    selected_index: SignalInstance = Signal(int, int)  # type: ignore
    object_placed: SignalInstance = Signal(ObjectButton)  # type: ignore

    def __init__(self, parent=None):
        super(TabbedToolBox, self).__init__(parent)

        self.setTabPosition(self.East)

        self._recent_toolbox = ObjectToolBox(self)
        self._recent_toolbox.object_icon_clicked.connect(self.selected)
        self._recent_toolbox.object_selected.connect(lambda index: self.selected_index.emit(0, index))
        self._recent_toolbox.object_placed.connect(self._on_object_dragged)

        self._objects_toolbox = ObjectToolBox(self)
        self._objects_toolbox.object_icon_clicked.connect(self.selected)
        self._objects_toolbox.object_selected.connect(lambda index: self.selected_index.emit(1, index))
        self._objects_toolbox.object_placed.connect(self._on_object_dragged)

        self._enemies_toolbox = ObjectToolBox(self)
        self._enemies_toolbox.object_icon_clicked.connect(self.selected)
        self._enemies_toolbox.object_selected.connect(lambda index: self.selected_index.emit(2, index))
        self._enemies_toolbox.object_placed.connect(self._on_object_dragged)

        self._recent_scroll_area = QScrollArea(self)
        self._recent_scroll_area.setWidgetResizable(True)
        self._recent_scroll_area.setWidget(self._recent_toolbox)

        self._object_scroll_area = QScrollArea(self)
        self._object_scroll_area.setWidgetResizable(True)
        self._object_scroll_area.setWidget(self._objects_toolbox)

        self._enemies_scroll_area = QScrollArea(self)
        self._enemies_scroll_area.setWidgetResizable(True)
        self._enemies_scroll_area.setWidget(self._enemies_toolbox)

        self.addTab(self._recent_scroll_area, "Recent")
        self.addTab(self._object_scroll_area, "Objects")
        self.addTab(self._enemies_scroll_area, "Enemies")

        self.show_level_object_tab()

        self.setWhatsThis(
            "<b>Object Toolbox</b><br/>"
            "Contains all objects and enemies/items, that can be placed in this type of level. Which are "
            "available depends on the object set, that is selected for this level.<br/>"
            "You can drag and drop objects into the level or click to select them. After selecting "
            "an object, you can place it by clicking the middle mouse button anywhere in the level."
            "<br/><br/>"
            "Note: Some items, like blocks with items in them, are displayed as they appear in the ROM, "
            "mouse over them and check their names in the ToolTip, or use the object dropdown to find "
            "them directly."
        )

    def sizeHint(self):
        size = super().sizeHint()
        width = self._recent_toolbox.sizeHint().width()
        width = max(width, self._objects_toolbox.sizeHint().width())
        width = max(width, self._enemies_toolbox.sizeHint().width())

        size.setWidth(
            max(width, size.width()) + self.tabBar().width() + self._object_scroll_area.verticalScrollBar().width() + 10
        )

        return size

    def show_recent_tab(self):
        self.setCurrentIndex(self.indexOf(self._recent_toolbox))

    def show_level_object_tab(self):
        self.setCurrentIndex(self.indexOf(self._object_scroll_area))

    def show_enemy_item_tab(self):
        self.setCurrentIndex(self.indexOf(self._enemies_scroll_area))

    def select_object(self, tab_index: int, object_index: int) -> Union[LevelObject, EnemyObject]:
        if tab_index == 0:
            return self._recent_toolbox.objects[object_index]
        elif tab_index == 1:
            return self._objects_toolbox.objects[object_index]
        else:
            return self._enemies_toolbox.objects[object_index]

    def set_object_set(self, object_set_index, graphic_set_index=-1):
        self._recent_toolbox.clear()
        self._objects_toolbox.clear()
        self._objects_toolbox.add_from_object_set(object_set_index, graphic_set_index)

        self._enemies_toolbox.clear()
        self._enemies_toolbox.add_from_enemy_set(object_set_index)

    def add_recent_object(self, level_object: Union[EnemyObject, LevelObject]):
        self._recent_toolbox.place_at_front(level_object)

    def _on_object_dragged(self, object_icon: ObjectButton):
        self.add_recent_object(object_icon.item)



from json import loads

from PySide6.QtCore import QRect

from foundry import warp_definitions
from foundry.core.Position import PositionProtocol
from foundry.game.Definitions import Definition
from foundry.game.gfx.objects.GeneratorObject import GeneratorObject
from foundry.game.gfx.objects.LevelObject import GROUND, SCREEN_HEIGHT, SCREEN_WIDTH


class Jump(GeneratorObject):
    POINTER_DOMAIN = 0b111

    def __init__(self, data):
        self.data = data

        self.blocks = []
        self.name = "Jump object"

        assert self.is_jump(data)

        self.screen_index = data[0] & 0x0F
        self.exit_vertical = (data[1] & 0xF0) >> 4
        self.exit_action = data[1] & 0x0F
        # for some reason those are flipped, meaning 5678, 1234
        self.exit_horizontal = ((data[2] & 0xF) << 4) + (data[2] >> 4)

    def to_bytes(self):
        return self.data

    def __repr__(self):
        return (
            f"Jump: Screen #{self.screen_index}, "
            + f"Exit ({self.exit_horizontal}, {self.exit_vertical}), "
            + f"Action #{self.exit_action}"
        )

    def __str__(self):
        return f"Jump on screen #{self.screen_index}"

    @staticmethod
    def is_jump(data):
        return data[0] >> 5 == Jump.POINTER_DOMAIN

    @staticmethod
    def from_properties(screen_index, action, horiz, vert):
        data = bytearray(3)

        data[0] |= 0b1110_0000
        data[0] |= screen_index

        data[1] |= vert << 4
        data[1] |= action

        data[2] |= ((horiz & 0xF) << 4) + (horiz >> 4)

        return Jump(data)

    def render(self):
        pass

    def draw(self, dc, zoom, transparent):
        pass

    def get_status_info(self):
        return []

    @property
    def definition(self) -> Definition:
        with open(warp_definitions, "r") as f:
            return Definition(__root__=loads(f.read()))

    @property
    def position(self) -> PositionProtocol:
        pass

    @position.setter
    def position(self, position: PositionProtocol) -> None:
        pass

    def move_by(self, dx, dy):
        pass

    def resize_by(self, dx, dy):
        pass

    def point_in(self, x, y):
        return False

    def change_type(self, new_type):
        pass

    def get_rect(self, block_length=1, vertical=False) -> QRect:
        if vertical:
            return QRect(
                0,
                block_length * (1 + SCREEN_HEIGHT * self.screen_index),
                block_length * SCREEN_WIDTH,
                block_length * SCREEN_HEIGHT,
            )
        else:
            return QRect(
                block_length * SCREEN_WIDTH * self.screen_index,
                0,
                block_length * SCREEN_WIDTH,
                block_length * GROUND,
            )

    def __contains__(self, point):
        return False



from enum import Enum


class WarningType(str, Enum):
    invalid_type = "INVALID OBJECT"
    invalid_position = "INVALID POSITION"
    invalid_size = "INVALID SIZE"
    invalid_extension_to_ground = "INVALID EXTENSION TO GROUND"
    outside_level_bounds = "OUTSIDE LEVEL BOUNDS"
    enemy_compatibility = "ENEMY COMPATIBILITY"
    invalid_warp = "INVALID WARP"

    @classmethod
    def has_value(cls, value: str) -> bool:
        """
        A convenience method to quickly determine if a value is a valid enumeration.

        Parameters
        ----------
        value : str
            The value to check against the enumeration.

        Returns
        -------
        bool
            If the value is inside the enumeration.
        """
        return value in cls._value2member_map_



from typing import Optional

from PySide6.QtCore import QObject, Signal, SignalInstance

from foundry.core.UndoController import UndoController
from foundry.game.level import LevelByteData
from foundry.game.level.Level import Level


class LevelRef(QObject):
    data_changed: SignalInstance = Signal()  # type: ignore
    jumps_changed: SignalInstance = Signal()  # type: ignore

    def __init__(self):
        super(LevelRef, self).__init__()
        self._internal_level = None
        self._undo_controller = None
        self._is_loaded = False

    @property
    def is_loaded(self) -> bool:
        return self._is_loaded

    def load_level(self, level_name: str, object_data_offset: int, enemy_data_offset: int, object_set_number: int):
        self.level = Level(level_name, object_data_offset, enemy_data_offset, object_set_number)
        self._is_loaded = True

        # actively emit, because we weren't connected yet, when the level sent it out
        self.data_changed.emit()

    def unload_level(self) -> None:
        self._internal_level = None
        self._undo_controller = None
        self._is_loaded = False

    @property
    def level(self) -> Level:
        assert self._internal_level is not None

        return self._internal_level

    @level.setter
    def level(self, level: Level):
        self._internal_level = level

        self._undo_controller = UndoController(self._internal_level.to_bytes())

        self._internal_level.data_changed.connect(self.data_changed.emit)
        self._internal_level.jumps_changed.connect(self.jumps_changed.emit)

    @property
    def selected_objects(self):
        assert self._internal_level is not None

        return [obj for obj in self._internal_level.get_all_objects() if obj.selected]

    @selected_objects.setter
    def selected_objects(self, selected_objects):
        assert self._internal_level is not None

        if selected_objects == self.selected_objects:
            return

        for obj in self._internal_level.get_all_objects():
            obj.selected = obj in selected_objects

        self.data_changed.emit()

    @property
    def state(self) -> LevelByteData:
        assert self._undo_controller is not None

        return self._undo_controller.state

    def do(self, level_data: Optional[LevelByteData] = None) -> LevelByteData:
        assert self._undo_controller is not None

        data = self._undo_controller.do(level_data if level_data is not None else self.level.to_bytes())
        self.data_changed.emit()
        return data

    @property
    def can_undo(self) -> bool:
        assert self._undo_controller is not None

        return self._undo_controller.can_undo

    def undo(self) -> LevelByteData:
        assert self._undo_controller is not None

        new_state = self._undo_controller.undo()
        self.set_level_state(*new_state)
        return new_state

    @property
    def can_redo(self) -> bool:
        assert self._undo_controller is not None

        return self._undo_controller.can_redo

    def redo(self) -> LevelByteData:
        assert self._undo_controller is not None

        new_state = self._undo_controller.redo()
        self.set_level_state(*new_state)
        return new_state

    def set_level_state(self, object_data, enemy_data):
        self.level.from_bytes(object_data, enemy_data, new_level=False)
        self.level.changed = True

        self.data_changed.emit()

    def save_level_state(self):
        assert self._internal_level is not None
        assert self._undo_controller is not None

        self.do(self._internal_level.to_bytes())
        self.level.changed = True

        self.data_changed.emit()

    def __bool__(self):
        return self.is_loaded



from typing import Optional

from attr import attrs
from PySide6.QtCore import Signal, SignalInstance
from PySide6.QtGui import QCloseEvent, QMouseEvent, QPainter, QPaintEvent, QResizeEvent
from PySide6.QtWidgets import QLayout, QStatusBar, QToolBar, QWidget

from foundry import icon
from foundry.core.Position import Position
from foundry.core.Size import Size, SizeProtocol
from foundry.core.sprites import SPRITE_SIZE
from foundry.game.gfx.drawable.Sprite import Sprite
from foundry.game.gfx.GraphicsSet import GraphicsSetProtocol
from foundry.game.gfx.Palette import PaletteGroupProtocol
from foundry.gui.CustomChildWindow import CustomChildWindow


@attrs(slots=True, auto_attribs=True)
class SpriteViewerModel:
    graphics_set: GraphicsSetProtocol
    palette_group: PaletteGroupProtocol
    palette_index: int


class SpriteViewerController(CustomChildWindow):
    graphics_set_changed: SignalInstance = Signal(GraphicsSetProtocol)  # type: ignore
    palette_group_changed: SignalInstance = Signal(PaletteGroupProtocol)  # type: ignore
    palette_index_changed: SignalInstance = Signal(int)  # type: ignore
    sprite_selected: SignalInstance = Signal(int)  # type: ignore
    destroyed: SignalInstance = Signal()  # type: ignore

    def __init__(
        self,
        parent: Optional[QWidget],
        graphics_set: GraphicsSetProtocol,
        palette_group: PaletteGroupProtocol,
        palette_index: int,
    ):
        super().__init__(parent, "Sprite Viewer")

        self.model = SpriteViewerModel(graphics_set, palette_group, palette_index)
        self.view = SpriteViewerView(self, graphics_set, palette_group, palette_index)
        self.setCentralWidget(self.view)
        self.toolbar = QToolBar(self)

        self.toolbar.setMovable(False)

        self.view.sprite_selected.connect(self.sprite_selected.emit)

        self.zoom_out_action = self.toolbar.addAction(icon("zoom-out.svg"), "Zoom Out")
        self.zoom_in_action = self.toolbar.addAction(icon("zoom-in.svg"), "Zoom In")

        def change_zoom(offset: int):
            self.view.zoom += offset

        self.zoom_out_action.triggered.connect(lambda *_: change_zoom(-1))  # type: ignore
        self.zoom_in_action.triggered.connect(lambda *_: change_zoom(1))  # type: ignore

        self.addToolBar(self.toolbar)
        self.setStatusBar(QStatusBar(self))

        self.layout().setSizeConstraint(QLayout.SetFixedSize)

    def closeEvent(self, event: QCloseEvent):
        self.toolbar.close()
        self.destroyed.emit()
        super().closeEvent(event)

    @property
    def graphics_set(self) -> GraphicsSetProtocol:
        return self.model.graphics_set

    @graphics_set.setter
    def graphics_set(self, value: GraphicsSetProtocol):
        self.model.graphics_set = value
        self.view.graphics_set = value
        self.graphics_set_changed.emit(self.graphics_set)
        self.view.update()

    @property
    def palette_group(self) -> PaletteGroupProtocol:
        return self.model.palette_group

    @palette_group.setter
    def palette_group(self, value: PaletteGroupProtocol):
        self.model.palette_group = value
        self.view.palette_group = value
        self.palette_group_changed.emit(self.palette_group)
        self.view.update()

    @property
    def palette_index(self) -> int:
        return self.model.palette_index

    @palette_index.setter
    def palette_index(self, value: int):
        self.model.palette_index = value
        self.view.palette_index = value
        self.palette_index_changed.emit(self.palette_index)
        self.view.update()


class SpriteViewerView(QWidget):
    sprite_selected: SignalInstance = Signal(int)  # type: ignore
    mouse_moved: SignalInstance = Signal(int, int, int, str)  # type: ignore

    SPRITES = 32
    SPRITES_PER_ROW = 16
    SPRITES_PER_COLUMN = 2

    def __init__(
        self,
        parent: Optional[QWidget],
        graphics_set: GraphicsSetProtocol,
        palette_group: PaletteGroupProtocol,
        palette_index: int,
        zoom: int = 4,
    ):
        super().__init__(parent)

        self.setMouseTracking(True)

        self.graphics_set = graphics_set
        self.palette_group = palette_group
        self.palette_index = palette_index
        self.zoom = zoom

    def resizeEvent(self, event: QResizeEvent):
        self.update()

    @property
    def zoom(self) -> int:
        return self._zoom

    @zoom.setter
    def zoom(self, value: int):
        self._zoom = value
        self.setFixedSize(
            self.SPRITES_PER_ROW * SPRITE_SIZE.width * self.zoom,
            self.SPRITES_PER_COLUMN * SPRITE_SIZE.height * self.zoom,
        )

    @property
    def sprite_size(self) -> SizeProtocol:
        return Size(SPRITE_SIZE.width * self.zoom, SPRITE_SIZE.height * self.zoom)

    def mouseMoveEvent(self, event: QMouseEvent):
        pos = Position.from_qpoint(event.pos())
        pos.x, pos.y = pos.x // self.sprite_size.width, pos.y // self.sprite_size.height

        dec_index = pos.y * self.SPRITES_PER_ROW + pos.x
        hex_index = hex(dec_index).upper().replace("X", "x")

        self.mouse_moved.emit(pos.x, pos.y, dec_index, hex_index)

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        pos = Position.from_qpoint(event.pos())
        pos.x, pos.y = pos.x // self.sprite_size.width, pos.y // self.sprite_size.height

        index = pos.y * self.SPRITES_PER_ROW + pos.x
        self.sprite_selected.emit(index)

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)

        for i in range(self.SPRITES):
            sprite = Sprite(
                i * 2,
                tuple(tuple(c for c in pal) for pal in self.palette_group),  # type: ignore
                self.palette_index,
                self.graphics_set,
            )

            x = (i % self.SPRITES_PER_ROW) * self.sprite_size.width
            y = (i // self.SPRITES_PER_ROW) * self.sprite_size.height

            sprite.draw(painter, x, y, SPRITE_SIZE.width * self.zoom, SPRITE_SIZE.height * self.zoom, transparent=True)



#!/usr/bin/env python3
import logging
import os
import sys
import traceback
from argparse import ArgumentParser, BooleanOptionalAction

import pretty_errors  # noqa: F401 is used to provide nicer messages throughout the project.
from PySide6.QtWidgets import QApplication, QMessageBox

from foundry import auto_save_rom_path, github_issue_link
from foundry.gui.AutoSaveDialog import AutoSaveDialog
from foundry.gui.settings import load_settings, save_settings

logger = logging.getLogger(__name__)

# change into the tmp directory pyinstaller uses for the data
if hasattr(sys, "_MEIPASS"):
    logger.info(f"Changing current dir to {getattr(sys, '_MEIPASS')}")
    os.chdir(getattr(sys, "_MEIPASS"))

from foundry.gui.MainWindow import MainWindow  # noqa: E402


def start():
    parser = ArgumentParser(description="The future of editing SMB3!")
    parser.add_argument("--path", dest="path", type=str, help="The path to the ROM", default="")
    parser.add_argument(
        "--dev", default=False, action=BooleanOptionalAction, type=bool, help="Override path with system path"
    )
    parser.add_argument("--level", type=int, help="Level index", default=None)
    parser.add_argument("--world", type=int, help="World Index", default=None)

    args = parser.parse_args()
    path: str = args.path
    if args.dev:
        dev_path = os.getenv("SMB3_TEST_ROM")
        if dev_path is not None:
            path = dev_path
    main(path, args.world, args.level)


def main(path_to_rom: str = "", world=None, level=None):
    load_settings()

    app = QApplication()

    if auto_save_rom_path.exists():
        result = AutoSaveDialog().exec_()

        if result == QMessageBox.AcceptRole:
            path_to_rom = auto_save_rom_path

            QMessageBox.information(
                None, "Auto Save recovered", "Don't forget to save the loaded ROM under a new name!"
            )

    window = MainWindow(path_to_rom, world, level)
    if window.loaded:
        del window.loaded
        app.exec_()
        save_settings()


if __name__ == "__main__":
    if len(sys.argv) > 1:
        path = sys.argv[1]
    else:
        path = ""

    try:
        main(path)
    except Exception as e:
        box = QMessageBox()
        box.setWindowTitle("Crash report")
        box.setText(
            f"An unexpected error occurred! Please contact the developers at {github_issue_link} "
            f"with the error below:\n\n{str(e)}\n\n{traceback.format_exc()}"
        )
        box.exec_()
        raise



from typing import Optional

from attr import attrs

from foundry.game.File import ROM
from foundry.game.gfx.Palette import (
    COLORS_PER_PALETTE,
    PALETTE_GROUPS_PER_OBJECT_SET,
    PALETTES_PER_PALETTES_GROUP,
    PaletteGroup,
    PaletteGroupProtocol,
    get_internal_palette_offset,
)


@attrs(slots=True, auto_attribs=True)
class PaletteGroupModel:
    tileset: int
    background_index: int
    sprite_index: int
    background_palette_group: PaletteGroupProtocol
    sprite_palette_group: PaletteGroupProtocol
    changed: bool = False
    background_palette_group_backup: Optional[PaletteGroupProtocol] = None
    sprite_palette_group_backup: Optional[PaletteGroupProtocol] = None

    def restore(self):
        self.background_palette_group = (
            self.background_palette_group_backup
            if self.background_palette_group_backup is not None
            else PaletteGroup.from_tileset(self.tileset, self.background_index)
        )
        self.sprite_palette_group = (
            self.sprite_palette_group_backup
            if self.sprite_palette_group_backup is not None
            else PaletteGroup.from_tileset(self.tileset, self.sprite_index + PALETTE_GROUPS_PER_OBJECT_SET)
        )

    def soft_save(self):
        if self.background_palette_group_backup is None:
            self.background_palette_group_backup = PaletteGroup.from_tileset(self.tileset, self.background_index)
        if self.sprite_palette_group_backup is None:
            self.sprite_palette_group_backup = PaletteGroup.from_tileset(
                self.tileset, self.sprite_index + PALETTE_GROUPS_PER_OBJECT_SET
            )
        self._save()

    def save(self, rom: Optional[ROM] = None):
        self.background_palette_group_backup = None
        self.sprite_palette_group_backup = None
        self._save(rom)

    def _save(self, rom: Optional[ROM] = None):
        bg_offset = (
            get_internal_palette_offset(self.tileset)
            + self.background_index * PALETTES_PER_PALETTES_GROUP * COLORS_PER_PALETTE
        )
        spr_offset = (
            get_internal_palette_offset(self.tileset)
            + (self.sprite_index + PALETTE_GROUPS_PER_OBJECT_SET) * PALETTES_PER_PALETTES_GROUP * COLORS_PER_PALETTE
        )

        if rom is None:
            rom = ROM()
        rom.write(bg_offset, bytes(self.background_palette_group))
        rom.write(spr_offset, bytes(self.sprite_palette_group))



from attr import attrs

from foundry.core.player_animations.PlayerAnimation import PlayerAnimation
from foundry.game.gfx.GraphicsSet import GraphicsSetProtocol
from foundry.game.gfx.Palette import PaletteGroupProtocol


@attrs(slots=True, auto_attribs=True)
class PlayerAnimationData:
    animation: PlayerAnimation
    graphics_set: GraphicsSetProtocol
    palette_group: PaletteGroupProtocol
    palette_index: int
    is_kicking: bool = False

    @property
    def horizontal_flip(self) -> list[bool]:
        if self.frames[3] == self.frames[4]:
            return [False, True, False, False, True, False]
        else:
            return [False, False, False, False, False, False]

    @property
    def frames(self) -> list[int]:
        animation = self.animation.animations
        if self.is_kicking:
            return [animation[0], animation[1], animation[2], animation[3], animation[4], animation[5]]
        return list(animation)



from json import loads
from typing import Optional

from pydantic import BaseModel

from foundry import data_dir


class Location(BaseModel):
    """
    A representation of a location of a level.

    world: int
        The world the level is located inside.
    index: int
        The index the level is inside the world.
    """

    world: int
    index: int


class DisplayInformation(BaseModel):
    """
    The display information to nicely sort levels.

    Attributes
    ----------
    name: Optional[str]
        The name of the level.
    description: Optional[str]
        The description of the level.
    locations: list[Location]
        The locations that the level is inside.
    """

    name: Optional[str]
    description: Optional[str]
    locations: list[Location]


class Level(BaseModel):
    """
    The representation of a level inside the game.

    Attributes
    ----------
    display_information: DisplayInformation
        Useful information regarding the level to make it human usable.
    generator_pointer: int
        The location this level's generators are located at.
    enemy_pointer: int
        The location this level's enemies are located at.
    tileset: int
        The tileset of the this level.
    generator_size: int
        The amount of space the generator data takes inside the game.
    enemy_size: int
        The amount of space the enemy data takes inside the game.
    """

    display_information: DisplayInformation
    generator_pointer: int
    enemy_pointer: int
    tileset: int
    generator_size: int
    enemy_size: int


def get_world_levels(world: int, levels: list[Level]) -> list[Level]:
    """
    Provides every level that is inside a given world.

    Parameters
    ----------
    world : int
        The world to select levels from.
    levels : list[Level]
        The index of levels to find levels from.

    Returns
    -------
    list[Level]
        A sub-list of levels which contains only the levels that were inside the given world.

    Notes
    -----
    For each level.display_information.index, there will only be a single level inside the return.
    """
    world_levels = {}
    for level in levels:
        for location in level.display_information.locations:
            if location.world == world:
                world_levels.update({location.index: level})
    return [element[1] for element in sorted(world_levels.items(), key=lambda item: item[0])]


def get_worlds(levels: list[Level]) -> int:
    """
    Determines the amount of worlds there are inside the game.

    Parameters
    ----------
    levels : list[Level]
        The levels to find worlds from.

    Returns
    -------
    int
        The amount of worlds there are.
    """
    worlds = 0
    for level in levels:
        if level.tileset == 0:
            worlds += 1
    return worlds


def load_level_offsets() -> list[Level]:
    with open(data_dir.joinpath("levels.json"), "r") as f:
        return [Level(**level) for level in loads(f.read())]



def little_endian(two_bytes: bytearray) -> int:
    """
    Takes a byte array of length 2 and returns the integer it represents in little endian.
    """

    first, second = two_bytes

    return (second << 8) + first



from collections import defaultdict
from typing import Any, Generator, List
from warnings import warn

from foundry.smb3parse.constants import (
    TILE_BOWSER_CASTLE,
    TILE_CASTLE_BOTTOM,
    TILE_DUNGEON_1,
    TILE_DUNGEON_2,
    TILE_HAND_TRAP,
    TILE_LEVEL_1,
    TILE_LEVEL_10,
    TILE_MUSHROOM_HOUSE_1,
    TILE_MUSHROOM_HOUSE_2,
    TILE_PIPE,
    TILE_POND,
    TILE_PYRAMID,
    TILE_QUICKSAND,
    TILE_SPADE_HOUSE,
    TILE_SPIRAL_TOWER_1,
    TILE_SPIRAL_TOWER_2,
    TILE_STAR_1,
    TILE_STAR_2,
)
from foundry.smb3parse.levels import (
    BASE_OFFSET,
    COMPLETABLE_LIST_END_MARKER,
    COMPLETABLE_TILES_LIST,
    ENEMY_BASE_OFFSET,
    FIRST_VALID_ROW,
    LAYOUT_LIST_OFFSET,
    LEVEL_ENEMY_LIST_OFFSET,
    LEVEL_X_POS_LISTS,
    LEVEL_Y_POS_LISTS,
    LEVELS_IN_WORLD_LIST_OFFSET,
    OFFSET_BY_OBJECT_SET_A000,
    OFFSET_SIZE,
    SPECIAL_ENTERABLE_TILE_AMOUNT,
    SPECIAL_ENTERABLE_TILES_LIST,
    STRUCTURE_DATA_OFFSETS,
    TILE_ATTRIBUTES_TS0_OFFSET,
    VALID_COLUMNS,
    VALID_ROWS,
    WORLD_COUNT,
    WORLD_MAP_BASE_OFFSET,
    WORLD_MAP_HEIGHT,
    WORLD_MAP_SCREEN_SIZE,
    WORLD_MAP_SCREEN_WIDTH,
    LevelBase,
)
from foundry.smb3parse.levels.level import Level
from foundry.smb3parse.objects.object_set import WORLD_MAP_OBJECT_SET
from foundry.smb3parse.util.rom import Rom

TILE_NAMES = defaultdict(lambda: "NO NAME")
TILE_NAMES.update(
    {
        TILE_MUSHROOM_HOUSE_1: "Mushroom House",
        TILE_MUSHROOM_HOUSE_2: "Mushroom House",
        TILE_SPIRAL_TOWER_1: "Spiral Tower",
        TILE_SPIRAL_TOWER_2: "Spiral Tower",
        TILE_DUNGEON_1: "Dungeon",
        TILE_DUNGEON_2: "Dungeon",
        TILE_QUICKSAND: "Quicksand",
        TILE_PYRAMID: "Pyramid",
        TILE_PIPE: "Pipe",
        TILE_POND: "Pond",
        TILE_CASTLE_BOTTOM: "Peach's Castle",
        TILE_BOWSER_CASTLE: "Bowser's Lair",
        TILE_HAND_TRAP: "Hand Trap",
        TILE_SPADE_HOUSE: "Spade Bonus",
        TILE_STAR_1: "Star",
        TILE_STAR_2: "Star",
    }
)


def list_world_map_addresses(rom: Rom) -> List[int]:
    offsets = rom.read(LAYOUT_LIST_OFFSET, WORLD_COUNT * OFFSET_SIZE)

    addresses = []

    for world in range(WORLD_COUNT):
        index = world * 2

        world_map_offset = (offsets[index + 1] << 8) + offsets[index]

        addresses.append(WORLD_MAP_BASE_OFFSET + world_map_offset)

    return addresses


def get_all_world_maps(rom: Rom) -> List["WorldMap"]:
    world_map_addresses = list_world_map_addresses(rom)

    return [WorldMap(address, rom) for address in world_map_addresses]


class WorldMap(LevelBase):
    """
    Represents the data associated with a world map/overworld. World maps are always 9 blocks high and 16 blocks wide.
    They can be multiple screens big, which are either not visibly connected or connected horizontally.

    Attributes:
        layout_address  The position in the ROM of the bytes making up the visual layout of the world map.
        layout_bytes    The actual bytes making up the visual layout

        width           The width of the world map in blocks across all scenes.
        height          The height of the world map, always 9 blocks.

        object_set      An ObjectSet object for the world map object set.
        screen_count    How many screens this world map spans.
    """

    def __init__(self, layout_address: int, rom: Rom):
        super(WorldMap, self).__init__(WORLD_MAP_OBJECT_SET, layout_address)

        self._rom = rom

        self._minimal_enterable_tiles = _get_normal_enterable_tiles(self._rom)
        self._special_enterable_tiles = _get_special_enterable_tiles(self._rom)
        self._completable_tiles = _get_completable_tiles(self._rom)

        memory_addresses = list_world_map_addresses(rom)

        try:
            self.number = memory_addresses.index(layout_address) + 1
        except ValueError:
            raise ValueError(f"World map was not found at given memory address {hex(layout_address)}.")

        self.height = WORLD_MAP_HEIGHT

        layout_end_index = rom.find(b"\xFF", layout_address)

        self.layout_bytes = rom.read(layout_address, layout_end_index - layout_address)

        if len(self.layout_bytes) % WORLD_MAP_SCREEN_SIZE != 0:
            raise ValueError(
                f"Invalid length of layout bytes for world map ({self.layout_bytes}). "
                f"Should be divisible by {WORLD_MAP_SCREEN_SIZE}."
            )

        self.screen_count = len(self.layout_bytes) // WORLD_MAP_SCREEN_SIZE
        self.width = int(self.screen_count * WORLD_MAP_SCREEN_WIDTH)

        self._parse_structure_data_block(rom)

    @property
    def world_index(self):
        return self.number - 1

    @property
    def level_count(self):
        return self.level_count_s1 + self.level_count_s2 + self.level_count_s3 + self.level_count_s4

    def _parse_structure_data_block(self, rom: Rom):
        structure_block_offset = rom.little_endian(STRUCTURE_DATA_OFFSETS + OFFSET_SIZE * self.world_index)

        self.structure_block_start = WORLD_MAP_BASE_OFFSET + structure_block_offset

        # the indexes into the y_pos list, where the levels for the n-th screen start
        y_pos_start_by_screen = rom.read(self.structure_block_start, 4)

        level_y_pos_list_start = WORLD_MAP_BASE_OFFSET + rom.little_endian(
            LEVEL_Y_POS_LISTS + OFFSET_SIZE * self.world_index
        )

        level_x_pos_list_start = WORLD_MAP_BASE_OFFSET + rom.little_endian(
            LEVEL_X_POS_LISTS + OFFSET_SIZE * self.world_index
        )

        level_y_pos_list_end = level_x_pos_list_start - level_y_pos_list_start

        self.level_count_s1 = y_pos_start_by_screen[1] - y_pos_start_by_screen[0]
        self.level_count_s2 = y_pos_start_by_screen[2] - y_pos_start_by_screen[1]
        self.level_count_s3 = y_pos_start_by_screen[3] - y_pos_start_by_screen[2]
        self.level_count_s4 = level_y_pos_list_end - y_pos_start_by_screen[3]

    def level_for_position(self, screen: int, player_row: int, player_column: int):
        """
        The rom takes the position of the current player, so the world, the screen and the x and y coordinates, and
        operates on them. First it is checked, whether or not the player is located on a tile, that is able to be
        entered.

        If that is the case, the x and y coordinates are used to look up the object set and address of the level. The
        object set is necessary to find the right base offset for the level address and to correctly parse its object
        data.

        Using the tile information it should be possible to correctly name almost all levels, for example Level *-1 will
        be located in the list at the offset pointed to by the "Level 1" tile in that world.

        That means, that all levels should be able to be collected, by iterating over all possible tiles and following
        the same procedure as the rom.

        :param screen:
        :param player_row:
        :param player_column:

        :return: A tuple of the object set number, the absolute level address, pointing to the objects and the enemy
        address. Or None, if there is no level at the map position.
        """
        assert isinstance(screen, int)
        assert isinstance(player_row, int)
        assert isinstance(player_column, int)

        tile = self.tile_at(screen, player_row, player_column)

        if tile in [TILE_SPADE_HOUSE, TILE_MUSHROOM_HOUSE_1, TILE_MUSHROOM_HOUSE_2]:
            warn("Spade and mushroom house currently not supported, when getting a level address.")
            return None

        if not self.is_enterable(tile):
            return None

        level_indexes = self.level_indexes(screen, player_row, player_column)

        if level_indexes is None:
            return None

        row_address, column_address, level_offset_address, enemy_offset_address = level_indexes

        level_offset = self._rom.little_endian(level_offset_address)

        assert 0xA000 <= level_offset < 0xC000, level_offset  # suppose that level layouts are only in this range?

        correct_row_value = self._rom.int(row_address)
        object_set_number = correct_row_value & 0x0F

        object_set_offset = (self._rom.int(OFFSET_BY_OBJECT_SET_A000 + object_set_number) * 2 - 10) * 0x1000

        absolute_level_address = 0x0010 + object_set_offset + level_offset

        # get enemy address
        enemy_address = ENEMY_BASE_OFFSET + self._rom.little_endian(enemy_offset_address)

        return object_set_number, absolute_level_address, enemy_address

    def replace_level_at_position(self, level_info, position):
        level_address, enemy_address, object_set_number = level_info

        existing_level = self.level_for_position(position.screen, position.row, position.column)

        if existing_level is None:
            raise LookupError("No existing level at position.")

        _, screen, row, column = position.tuple()

        row_address, column_address, level_offset_address, enemy_offset_address = self.level_indexes(
            screen, row, column
        )

        row_value = ((row + FIRST_VALID_ROW) << 4) + object_set_number
        self._rom.write(row_address, bytes([row_value]))

        column_value = ((screen - 1) << 4) + column
        self._rom.write(column_address, bytes([column_value]))

        object_set_offset = (self._rom.int(OFFSET_BY_OBJECT_SET_A000 + object_set_number) * 2 - 10) * 0x1000
        level_offset = level_address - object_set_offset - BASE_OFFSET

        self._rom.write_little_endian(level_offset_address, level_offset)

        enemy_offset = enemy_address - BASE_OFFSET

        self._rom.write_little_endian(enemy_offset_address, enemy_offset)

    def level_indexes(self, screen, player_row, player_column):
        """

        :param int screen: On which screen the level is positioned.
        :param int player_row: In which row the level is positioned.
        :param int player_column: In which column the level is positioned.

        :return: The memory addresses of the row, column and level offset position.
        """

        level_y_pos_list_start = WORLD_MAP_BASE_OFFSET + self._rom.little_endian(
            LEVEL_Y_POS_LISTS + OFFSET_SIZE * self.world_index
        )

        level_x_pos_list_start = WORLD_MAP_BASE_OFFSET + self._rom.little_endian(
            LEVEL_X_POS_LISTS + OFFSET_SIZE * self.world_index
        )

        row_amount = col_amount = level_x_pos_list_start - level_y_pos_list_start

        row_start_index = sum(
            [self.level_count_s1, self.level_count_s2, self.level_count_s3, self.level_count_s4][0 : screen - 1]
        )

        # find the row position
        for row_index in range(row_start_index, row_amount):
            value = self._rom.int(level_y_pos_list_start + row_index)

            # adjust the value, so that we ignore the black border tiles around the map
            row = (value >> 4) - FIRST_VALID_ROW

            if row == player_row:
                break
        else:
            # no level on row of player
            return None

        for col_index in range(row_index, col_amount):
            column = self._rom.int(level_x_pos_list_start + col_index) & 0x0F

            if column == player_column:
                break
        else:
            # no column for row
            return None

        row_position = level_y_pos_list_start + col_index
        col_position = level_x_pos_list_start + col_index

        # get level offset
        level_list_offset_position = LEVELS_IN_WORLD_LIST_OFFSET + self.world_index * OFFSET_SIZE
        level_list_address = WORLD_MAP_BASE_OFFSET + self._rom.little_endian(level_list_offset_position)

        level_offset_position = level_list_address + OFFSET_SIZE * col_index

        enemy_list_start_offset = LEVEL_ENEMY_LIST_OFFSET + self.world_index * OFFSET_SIZE
        enemy_list_start = WORLD_MAP_BASE_OFFSET + self._rom.little_endian(enemy_list_start_offset)

        enemy_offset_position = enemy_list_start + col_index * OFFSET_SIZE

        return row_position, col_position, level_offset_position, enemy_offset_position

    def level_name_for_position(self, screen: int, player_row: int, player_column: int) -> str:
        tile = self.tile_at(screen, player_row, player_column)

        if not self.is_enterable(tile):
            return ""

        if tile in range(TILE_LEVEL_1, TILE_LEVEL_10 + 1):
            return f"Level {self.number}-{tile - TILE_LEVEL_1 + 1}"

        return f"Level {self.number}-{TILE_NAMES[tile]}"

    def tile_at(self, screen: int, row: int, column: int) -> int:
        """
        Returns the tile value at the given coordinates. We define (0, 0) to be the topmost, leftmost tile, under the
        black border, so we'll adjust them accordingly, when bound checking.

        :param screen:
        :param row:
        :param column:
        :return:
        """
        assert isinstance(screen, int)
        assert isinstance(row, int)
        assert isinstance(column, int)

        if row + FIRST_VALID_ROW not in VALID_ROWS:
            raise ValueError(
                f"Given row {row} is outside the valid range for world maps. First valid row is " f"{FIRST_VALID_ROW}."
            )

        if column not in VALID_COLUMNS:
            raise ValueError(
                f"Given column {column} is outside the valid range for world maps. Remember the black " f"border."
            )

        if screen - 1 not in range(self.screen_count):
            raise ValueError(f"World {self.number} has {self.screen_count} screens. " f"Given number {screen} invalid.")

        return self.layout_bytes[(screen - 1) * WORLD_MAP_SCREEN_SIZE + row * WORLD_MAP_SCREEN_WIDTH + column]

    def is_enterable(self, tile_index: int) -> bool:
        """
        The tile attributes for the overworld tile set define the minimal value a tile has to have to be enterable.
        Which of the 4 bytes to check against depends on the "quadrant", so the 2 MSBs.

        :param tile_index: Tile index to check.

        :return: Whether the tile is enterable.
        """
        quadrant_index = tile_index >> 6

        # todo allows spade houses, but those break. treat them differently when loading their level
        return (
            tile_index >= self._minimal_enterable_tiles[quadrant_index]
            or tile_index in self._completable_tiles
            or tile_index in self._special_enterable_tiles
        )

    def gen_positions(self) -> Generator[Any, None, None]:
        """
        Returns a generator, which yield WorldMapPosition objects, one screen at a time, one row at a time.
        """
        from foundry.smb3parse.levels.WorldMapPosition import WorldMapPosition

        for screen in range(1, self.screen_count + 1):
            for row in range(WORLD_MAP_HEIGHT):
                for column in range(WORLD_MAP_SCREEN_WIDTH):
                    yield WorldMapPosition(self, screen, row, column)

    def gen_levels(self):
        """
        Returns a generator, which yields all levels accessible from this world map.
        """
        for position in self.gen_positions():
            level_info_tuple = self.level_for_position(position.screen, position.row, position.column)

            if level_info_tuple is None:
                continue

            else:
                yield Level(self._rom, *level_info_tuple)

    @staticmethod
    def from_world_number(rom: Rom, world_number: int) -> "WorldMap":
        if not world_number - 1 in range(WORLD_COUNT):
            raise ValueError(f"World number must be between 1 and {WORLD_COUNT}, including.")

        memory_address = list_world_map_addresses(rom)[world_number - 1]

        return WorldMap(memory_address, rom)

    def __repr__(self):
        return f"World {self.number}"


def _get_normal_enterable_tiles(rom: Rom) -> bytearray:
    return rom.read(TILE_ATTRIBUTES_TS0_OFFSET, 4)


def _get_special_enterable_tiles(rom: Rom) -> bytearray:
    return rom.read(SPECIAL_ENTERABLE_TILES_LIST, SPECIAL_ENTERABLE_TILE_AMOUNT)


def _get_completable_tiles(rom: Rom) -> bytearray:
    completable_tile_amount = rom.find(COMPLETABLE_LIST_END_MARKER, COMPLETABLE_TILES_LIST) - COMPLETABLE_TILES_LIST

    return rom.read(COMPLETABLE_TILES_LIST, completable_tile_amount)



from typing import Optional

from PySide6.QtCore import Signal, SignalInstance
from PySide6.QtWidgets import QVBoxLayout, QWidget

from foundry.game.gfx.Palette import PaletteGroupProtocol
from foundry.gui.PaletteEditorWidget import PaletteEditorWidget


class PaletteGroupEditor(QWidget):
    palette_group_changed: SignalInstance = Signal(PaletteGroupProtocol)  # type: ignore

    def __init__(self, parent: Optional[QWidget], palette_group: PaletteGroupProtocol):
        super().__init__(parent)
        self._palette_group = palette_group

        layout = QVBoxLayout()
        layout.setSpacing(0)

        self._palettes: list[PaletteEditorWidget] = []
        for idx, palette in enumerate(palette_group.palettes):
            widget = PaletteEditorWidget(self, palette)
            widget.palette_changed.connect(lambda *_, idx=idx: self._on_palette_changed(idx))
            self._palettes.append(widget)
            layout.addWidget(widget)

        self.setLayout(layout)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.parent}, {self.palette_group})"

    @property
    def palette_group(self) -> PaletteGroupProtocol:
        return self._palette_group

    @palette_group.setter
    def palette_group(self, palette_group: PaletteGroupProtocol):
        self._palette_group = palette_group
        self.palette_group_changed.emit(palette_group)
        self._update()

    def _update(self):
        for idx, palette in enumerate(self._palettes):
            palette._palette = self._palette_group[idx]
            palette._update()

    def _on_palette_changed(self, palette_index: int):
        palette_group = self.palette_group
        palette_group.palettes[palette_index] = self._palettes[palette_index].palette
        for palette in palette_group:
            palette[0] = palette_group[palette_index][0]
        self.palette_group = palette_group



import logging
import os
from collections.abc import Callable
from typing import Optional, Protocol, Tuple

from PySide6.QtCore import QPoint
from PySide6.QtGui import QAction, QPixmap, Qt
from PySide6.QtWidgets import QDialog

from foundry.core.Data import Data, DataProtocol
from foundry.game.File import ROM
from foundry.game.gfx.objects.EnemyItem import EnemyObject
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.level.Level import Level, get_level_name_suggestion
from foundry.game.level.LevelControlled import LevelControlled
from foundry.game.level.LevelRef import LevelRef
from foundry.gui.AutoScrollEditor import AutoScrollEditor
from foundry.gui.BlockViewer import BlockViewerController as BlockViewer
from foundry.gui.HeaderEditor import HeaderEditor
from foundry.gui.JumpEditor import JumpEditor
from foundry.gui.LevelSelector import LevelSelector
from foundry.gui.LevelView import undoable
from foundry.gui.PaletteViewer import PaletteViewer


def require_safe_to_change(function: Callable):
    def required_safe_to_change_wrapper(self, *args, **kwargs):
        if self.parent.safe_to_change:
            return function(self, *args, **kwargs)

    return required_safe_to_change_wrapper


class LevelSelectorController(Protocol):
    last_level: Optional[tuple[int, int]]


class LevelController:
    def __init__(self, parent: LevelControlled, level_ref: LevelRef):
        self.parent = parent
        self.level_ref = level_ref
        self.level_ref.data_changed.connect(self.on_level_data_changed)
        self.level_selector_last_level = None

    @property
    def changes(self) -> list[DataProtocol]:
        (gen_loc, gen_data), (ene_loc, ene_data) = self.level_ref.level.to_bytes()
        return [Data(gen_loc, gen_data), Data(ene_loc, ene_data)]

    @property
    def screenshot(self) -> QPixmap:
        return self.parent.level_view.make_screenshot()

    @property
    def title_suggestion(self) -> str:
        return f"{self.parent.level_view.level_ref.level.name} - {ROM.name}"

    @property
    def last_position(self) -> tuple[int, int]:
        return self.parent.context_menu.get_position()

    @property
    def actions(self) -> list[QAction]:
        return self.parent.context_menu.get_all_menu_item_ids()

    @property
    def suggested_object(self) -> int:
        return self.parent.object_dropdown.currentIndex()

    @require_safe_to_change
    def on_reload(self) -> None:
        self.update_level(
            level_name=self.level_ref.level.name,
            object_data_offset=self.level_ref.level.header_offset,
            enemy_data_offset=self.level_ref.level.enemy_offset,
            object_set=self.level_ref.level.object_set_number,
        )

    @require_safe_to_change
    def warp_to_alternative(self):
        level_address = self.level_ref.level.next_area_objects
        enemy_address = self.level_ref.level.next_area_enemies + 1
        object_set = self.level_ref.level.next_area_object_set

        self.update_level(
            f"Level {get_level_name_suggestion(level_address + 9)}", level_address, enemy_address, object_set
        )

    @require_safe_to_change
    def on_select(self):
        selector = LevelSelector(self.parent, start_level=self.level_selector_last_level)

        if QDialog.Accepted == selector.exec():
            self.level_selector_last_level = selector.current_level_index
            self.update_level(
                selector.level_name if selector.level_name is not None else "",
                selector.object_data_offset,
                selector.enemy_data_offset,
                selector.object_set,
            )

    def to_data(self) -> list[DataProtocol]:
        (object_offset, object_data), (enemy_offset, enemy_data) = self.level_ref.state

        return [Data(object_offset, object_data), Data(enemy_offset, enemy_data)]

    def load_m3l(self, m3l_data: bytearray, pathname: str):
        self.level_ref.level.from_m3l(m3l_data)
        self.level_ref.level.name = os.path.basename(pathname)

        self.update_gui_for_level()

    def to_m3l(self) -> bytearray:
        return self.level_ref.level.to_m3l()

    def refresh(self):
        self.on_reload()
        self.level_ref.level.reload()

    def update(self):
        self.parent.level_view.update()

    @property
    def safe_to_change(self) -> bool:
        if not self.level_ref.is_loaded:
            return True
        return not self.level_ref.level.changed and not self.parent.side_palette.changed

    @safe_to_change.setter
    def safe_to_change(self, value: bool):
        self.level_ref.level.changed = not value
        self.parent.side_palette.changed = not value

    @property
    def stable_changes(self) -> tuple[bool, str, str]:
        return self.parent.level_view.level_safe_to_save()

    @property
    def is_attached(self) -> bool:
        return self.level_ref.level.attached_to_rom

    @is_attached.setter
    def is_attached(self, value: bool):
        self.level_ref.level.attached_to_rom = value

    def attach(self, generator_pointer: int, enemy_pointer: int):
        self.level_ref.level.attach_to_rom(generator_pointer, enemy_pointer)

    def zoom_in(self):
        self.parent.level_view.zoom_in()

    def zoom_out(self):
        self.parent.level_view.zoom_out()

    def middle_mouse_release(self, position: QPoint):
        pointf = self.parent.level_view.mapFromGlobal(position)
        pos = (int(pointf.x()), int(pointf.y()))

        # If the number is negative do not do anything
        if any([num < 0 for num in pos]):
            return
        self.place_object_from_dropdown(pos)

    @undoable
    def to_foreground(self):
        self.level_ref.level.bring_to_foreground(self.level_ref.selected_objects)

    @undoable
    def to_background(self):
        self.level_ref.level.bring_to_background(self.level_ref.selected_objects)

    @undoable
    def create_object_at(self, x: int, y: int):
        self.parent.level_view.create_object_at(x, y)

    @undoable
    def create_enemy_at(self, x: int, y: int):
        self.parent.level_view.create_enemy_at(x, y)

    def cut(self):
        self.copy()
        self.delete()

    def copy(self):
        if selected := self.parent.level_view.get_selected_objects().copy():
            self.parent.context_menu.set_copied_objects(selected)

    @undoable
    def paste(self, x: Optional[int] = None, y: Optional[int] = None):
        self.parent.level_view.paste_objects_at(self.parent.context_menu.get_copied_objects(), x, y)

    def select_all(self):
        self.parent.level_view.select_all()
        self.parent.object_list.selectAll()

    def focus_selected(self):
        self.parent.object_dropdown.setFocus()

    @undoable
    def delete(self):
        self.parent.jump_list.delete()
        self.parent.level_view.remove_selected_objects()
        self.parent.level_view.update()
        self.parent.spinner_panel.disable_all()

    @undoable
    def on_jump_added(self):
        self.parent.level_view.add_jump()

    @undoable
    def on_jump_removed(self):
        self.parent.level_view.remove_jump(self.parent.jump_list.currentIndex().row())

    @undoable
    def on_jump_edited(self, jump):
        index = self.parent.jump_list.currentIndex().row()

        assert index >= 0

        if isinstance(self.level_ref.level, Level):
            self.level_ref.level.jumps[index] = jump
            self.parent.jump_list.item(index).setText(str(jump))

    @undoable
    def place_object_from_dropdown(self, pos: Tuple[int, int]) -> None:
        # the dropdown is synchronized with the toolbar, so it doesn't matter where to take it from
        level_object = self.parent.object_dropdown.currentData(Qt.UserRole)

        self.parent.object_toolbar.add_recent_object(level_object)

        if isinstance(level_object, LevelObject):
            self.parent.level_view.create_object_at(*pos, level_object.domain, level_object.obj_index)
        elif isinstance(level_object, EnemyObject):
            self.parent.level_view.add_enemy(level_object.obj_index, *pos, -1)

    @undoable
    def on_spin(self, _):
        selected_objects = self.level_ref.selected_objects

        if len(selected_objects) != 1:
            logging.error(selected_objects, RuntimeWarning)
            return

        selected_object = selected_objects[0]

        obj_type = self.parent.spinner_panel.get_type()

        if isinstance(selected_object, LevelObject):
            domain = self.parent.spinner_panel.get_domain()

            if selected_object.is_4byte:
                length = self.parent.spinner_panel.get_length()
            else:
                length = None

            self.parent.level_view.replace_object(selected_object, domain, obj_type, length)
        else:
            self.parent.level_view.replace_enemy(selected_object, obj_type)

        self.level_ref.data_changed.emit()

    def on_level_data_changed(self):
        self.parent.undo_action.setEnabled(self.level_ref.can_undo)
        self.parent.redo_action.setEnabled(self.level_ref.can_redo)

        self.parent.jump_destination_action.setEnabled(self.level_ref.level.has_next_area)
        self.parent.menu_toolbar_save_action.setEnabled(
            self.level_ref.level.changed or not self.level_ref.level.attached_to_rom or self.parent.side_palette.changed
        )

    def update_level(self, level_name: str, object_data_offset: int, enemy_data_offset: int, object_set: int):
        self.level_ref.load_level(level_name, object_data_offset, enemy_data_offset, object_set)
        self.update_gui_for_level()
        self.parent.update_title()
        self.parent.side_palette.load_from_level(self.level_ref.level)

    def update_gui_for_level(self):
        self.parent.side_palette.load_from_level(self.level_ref.level)

        self.parent.jump_list.update()
        self.parent.object_dropdown.set_object_set(
            self.level_ref.level.object_set_number, self.level_ref.level.graphic_set
        )
        self.parent.object_toolbar.set_object_set(
            self.level_ref.level.object_set_number, self.level_ref.level.graphic_set
        )

        self.parent.level_view.update()

    def display_header_editor(self):
        header_editor = HeaderEditor(self.parent, self.level_ref)  # type: ignore
        header_editor.tab_widget.setCurrentIndex(3)
        header_editor.exec_()

    def display_jump_editor(self):
        index = self.parent.jump_list.currentIndex().row()

        updated_jump = JumpEditor.edit_jump(
            self.parent,  # type: ignore
            self.level_ref.level.jumps[index],
            not self.level_ref.level.is_vertical,
            (self.level_ref.level.length // 0x10) - 1,
        )

        self.on_jump_edited(updated_jump)

    def display_palette_viewer(self):
        PaletteViewer(self.parent, self.level_ref).exec_()

    def display_autoscroll_editor(self):
        AutoScrollEditor(self.parent, self.level_ref).exec_()

    def display_block_viewer(self):
        block_viewer = BlockViewer(parent=self.parent)
        block_viewer.tileset = self.level_ref.level.object_set.number
        block_viewer.palette_group = self.level_ref.level.object_palette_index
        block_viewer.tile_square_assembly_changed.connect(self.update)
        block_viewer.show()

    def display_warnings(self):
        self.parent.warning_list.show()



from typing import List, Optional

from PySide6.QtWidgets import QCheckBox, QLabel, QVBoxLayout

from foundry.game.gfx.objects.EnemyItem import EnemyObject
from foundry.game.level.LevelRef import LevelRef
from foundry.gui.CustomDialog import CustomDialog
from foundry.gui.Spinner import Spinner
from foundry.smb3parse.constants import OBJ_AUTOSCROLL

AUTOSCROLL_LABELS = {
    -1: "No Autoscroll in Level.",
    0: "Horizontal Autoscroll",
    1: "Horizontal Autoscroll",
    2: "Moves Level up and right; screen wraps, vertically",
    3: "Moves ceiling down and up (Fortress Spike Levels)",
    4: "Moves ground up, until a door hits the ground",
    5: "Moves ground up and down, used for changes in over-water Levels",
}


class AutoScrollEditor(CustomDialog):
    def __init__(self, parent, level_ref: LevelRef):
        super(AutoScrollEditor, self).__init__(parent, title="Autoscroll Editor")
        self.level_ref = level_ref

        self.original_autoscroll_item = _get_autoscroll(self.level_ref.level.enemies)

        QVBoxLayout(self)

        self.enabled_checkbox = QCheckBox("Enable Autoscroll in level", self)
        self.enabled_checkbox.toggled.connect(self._insert_autoscroll_object)

        self.y_position_spinner = Spinner(self, maximum=0x60 - 1)
        self.y_position_spinner.valueChanged.connect(self._update_y_position)

        self.auto_scroll_type_label = QLabel(self)

        self.layout().addWidget(self.enabled_checkbox)
        self.layout().addWidget(self.y_position_spinner)
        self.layout().addWidget(self.auto_scroll_type_label)

        self.update()

    def update(self):
        autoscroll_item = _get_autoscroll(self.level_ref.level.enemies)

        self.enabled_checkbox.setChecked(autoscroll_item is not None)
        self.y_position_spinner.setEnabled(autoscroll_item is not None)

        if autoscroll_item is None:
            self.auto_scroll_type_label.setText(AUTOSCROLL_LABELS[-1])
        else:
            self.y_position_spinner.setValue(autoscroll_item.position.y)
            self.auto_scroll_type_label.setText(AUTOSCROLL_LABELS[autoscroll_item.position.y >> 4])

    def _update_y_position(self, _):
        autoscroll_item = _get_autoscroll(self.level_ref.level.enemies)

        if autoscroll_item is not None:
            autoscroll_item.position.y = self.y_position_spinner.value()

        self.level_ref.data_changed.emit()

        self.update()

    def _insert_autoscroll_object(self, should_insert: bool):
        autoscroll_item = _get_autoscroll(self.level_ref.level.enemies)

        if autoscroll_item is not None:
            self.level_ref.level.enemies.remove(autoscroll_item)

        if should_insert:
            self.level_ref.level.enemies.insert(0, self._create_autoscroll_object())

        self.level_ref.data_changed.emit()

        self.update()

    def _create_autoscroll_object(self):
        return self.level_ref.level.enemy_item_factory.from_properties(
            OBJ_AUTOSCROLL, 0, self.y_position_spinner.value()
        )

    def closeEvent(self, event):
        current_autoscroll_item = _get_autoscroll(self.level_ref.level.enemies)

        if not (self.original_autoscroll_item is None and current_autoscroll_item is None):
            added_or_removed_autoscroll = self.original_autoscroll_item is None or current_autoscroll_item is None

            if (
                added_or_removed_autoscroll
                or self.original_autoscroll_item.position.y != current_autoscroll_item.position.y  # type: ignore
            ):
                self.level_ref.save_level_state()

        super(AutoScrollEditor, self).closeEvent(event)


def _get_autoscroll(enemy_items: List[EnemyObject]) -> Optional[EnemyObject]:
    for item in enemy_items:
        if item.obj_index == OBJ_AUTOSCROLL:
            return item
    else:
        return None



from enum import Enum
from typing import List, Tuple, Union

from PySide6.QtCore import QPoint
from PySide6.QtWidgets import QMenu

from foundry.game.gfx.objects.EnemyItem import EnemyObject
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.level.LevelRef import LevelRef


class CMAction(Enum):
    REMOVE = 1
    ADD_OBJECT = 2
    COPY = 4
    PASTE = 5
    CUT = 6
    BACKGROUND = 7
    FOREGROUND = 8


class CMMode(Enum):
    BG = 1
    OBJ = 2
    LIST = 3


ID_PROP: bytes = "ID"

MAX_ORIGIN = 0xFF, 0xFF


class ContextMenu(QMenu):
    def __init__(self, level_ref: LevelRef):
        super(ContextMenu, self).__init__()

        self.level_ref = level_ref

        self.copied_objects = None
        self.copied_objects_origin = 0, 0
        self.last_opened_at = QPoint(0, 0)

        self.cut_action = self.addAction("Cut")
        self.cut_action.setProperty(ID_PROP, CMAction.CUT)

        self.copy_action = self.addAction("Copy")
        self.copy_action.setProperty(ID_PROP, CMAction.COPY)

        self.paste_action = self.addAction("Paste")
        self.paste_action.setProperty(ID_PROP, CMAction.PASTE)

        self.addSeparator()

        self.into_background_action = self.addAction("To Background")
        self.into_background_action.setProperty(ID_PROP, CMAction.BACKGROUND)

        self.into_foreground_action = self.addAction("To Foreground")
        self.into_foreground_action.setProperty(ID_PROP, CMAction.FOREGROUND)

        self.addSeparator()

        self.remove_action = self.addAction("Remove")
        self.remove_action.setProperty(ID_PROP, CMAction.REMOVE)

        self.add_object_action = self.addAction("Add Object")
        self.add_object_action.setProperty(ID_PROP, CMAction.ADD_OBJECT)

    def set_copied_objects(self, objects: List[Union[LevelObject, EnemyObject]]):
        if not objects:
            return

        self.copied_objects = objects

        min_x, min_y = MAX_ORIGIN

        for obj in objects:
            obj_x, obj_y = obj.position.x, obj.position.y

            min_x = min(min_x, obj_x)
            min_y = min(min_y, obj_y)

        min_x = max(min_x, 0)
        min_y = max(min_y, 0)

        self.copied_objects_origin = min_x, min_y

    def get_copied_objects(self) -> Tuple[List[Union[LevelObject, EnemyObject]], Tuple[int, int]]:
        return self.copied_objects, self.copied_objects_origin

    def set_position(self, position: QPoint):
        self.last_opened_at = position

    def get_position(self) -> Tuple[int, int]:
        x, y = self.last_opened_at.toTuple()

        return x, y

    def get_all_menu_item_ids(self):
        return [action.property("ID") for action in self.actions()]

    def as_object_menu(self) -> "ContextMenu":
        return self._setup_items(CMMode.OBJ)

    def as_background_menu(self) -> "ContextMenu":
        return self._setup_items(CMMode.BG)

    def as_list_menu(self) -> "ContextMenu":
        return self._setup_items(CMMode.LIST)

    def _setup_items(self, mode: CMMode):
        objects_selected = bool(self.level_ref.selected_objects)
        objects_copied = bool(self.copied_objects)

        self.cut_action.setEnabled(not mode == CMMode.BG and objects_selected)
        self.copy_action.setEnabled(not mode == CMMode.BG and objects_selected)
        self.paste_action.setEnabled(not mode == CMMode.LIST and objects_copied)

        self.into_background_action.setEnabled(not mode == CMMode.BG and objects_selected)
        self.into_foreground_action.setEnabled(not mode == CMMode.BG and objects_selected)

        self.remove_action.setEnabled(not mode == CMMode.BG and objects_selected)
        self.add_object_action.setEnabled(not mode == CMMode.LIST)

        return self



from json import loads
from typing import Optional

from PySide6.QtCore import Signal, SignalInstance
from PySide6.QtWidgets import QSizePolicy, QWidget

from foundry import spinner_panel_flags_path
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.gfx.objects.ObjectLike import ObjectLike
from foundry.game.level.LevelRef import LevelRef
from foundry.gui.util import setup_description, setup_layout

MAX_DOMAIN = 0x07
MAX_TYPE = 0xFF
MAX_LENGTH = 0xFF


class SpinnerPanel(QWidget):
    object_change: SignalInstance = Signal(int)

    zoom_in_triggered: SignalInstance = Signal()
    zoom_out_triggered: SignalInstance = Signal()

    def __init__(self, parent: Optional[QWidget], level_ref: LevelRef):
        super(SpinnerPanel, self).__init__(parent)

        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        self.level_ref = level_ref
        self.level_ref.data_changed.connect(self.update)

        with open(spinner_panel_flags_path, "r") as data:
            flags = loads(data.read())
        setup_layout(self, flags)
        setup_description(self, flags)

    def update(self):
        if len(self.level_ref.selected_objects) == 1:
            selected_object = self.level_ref.selected_objects[0]

            if isinstance(selected_object, ObjectLike):
                self._populate_spinners(selected_object)

        else:
            self.disable_all()

        super(SpinnerPanel, self).update()

    def _populate_spinners(self, obj: ObjectLike):
        self.blockSignals(True)

        self.set_type(obj.obj_index)

        if isinstance(obj, LevelObject):
            self.enable_domain(True, obj.domain)
        else:
            self.enable_domain(False)

        if isinstance(obj, LevelObject) and obj.is_4byte:
            self.set_length(obj.length)
        else:
            self.enable_length(False)

        self.blockSignals(False)

    def get_type(self):
        return self.spin_type.value()

    def set_type(self, object_type: int):
        self.spin_type.setValue(object_type)
        self.spin_type.setEnabled(True)

    def get_domain(self):
        return self.spin_domain.value()

    def set_domain(self, domain: int):
        self.spin_domain.setValue(domain)
        self.spin_domain.setEnabled(True)

    def get_length(self) -> int:
        return self.spin_length.value()

    def set_length(self, length: int):
        self.spin_length.setValue(length)
        self.spin_length.setEnabled(True)

    def enable_type(self, enable: bool, value: int = 0):
        self.spin_type.setValue(value)
        self.spin_type.setEnabled(enable)

    def enable_domain(self, enable: bool, value: int = 0):
        self.spin_domain.setValue(value)
        self.spin_domain.setEnabled(enable)

    def enable_length(self, enable: bool, value: int = 0):
        self.spin_length.setValue(value)
        self.spin_length.setEnabled(enable)

    def clear_spinners(self):
        self.set_type(0x00)
        self.set_domain(0x00)
        self.set_length(0x00)

    def disable_all(self):
        self.blockSignals(True)

        self.clear_spinners()

        self.enable_type(False)
        self.enable_domain(False)
        self.enable_length(False)

        self.blockSignals(False)



from foundry.core.warnings.Warning import PydanticWarning, Warning
from foundry.game.gfx.objects.EnemyItem import EnemyObject
from foundry.game.gfx.objects.ObjectLike import ObjectLike


def find_incompatibilities(enemies: list[EnemyObject]) -> tuple[bool, bool]:
    """
    Determines if there are any incompatibilities with enemy graphics.

    Parameters
    ----------
    enemies : list[EnemyObject]
        The enemies to check compatibility of.

    Returns
    -------
    tuple[bool, bool]
        If there are any incompatibilities between the 5th and 6th page, respectively.
    """
    primary_incompatibilities = set()
    secondary_incompatibilities = set()
    if not enemies:
        return False, False
    for enemy in enemies[:-1]:
        pages = enemy.definition.pages
        if len(pages) >= 3 and pages[2] != 0:
            primary_incompatibilities.add(pages[2])
        if len(pages) >= 4 and pages[3] != 0:
            secondary_incompatibilities.add(pages[3])
    last_def = enemies[-1].definition
    has_primary_incompatibilities = (
        bool(primary_incompatibilities)
        and len(last_def.pages) >= 3
        and last_def.pages[2] not in primary_incompatibilities
        and last_def.pages[2] != 0
    )
    has_secondary_incompatibilities = (
        bool(secondary_incompatibilities)
        and len(last_def.pages) >= 4
        and last_def.pages[3] not in secondary_incompatibilities
        and last_def.pages[3] != 0
    )
    return has_primary_incompatibilities, has_secondary_incompatibilities


class EnemyCompatibilityWarning(Warning):
    """
    A warning to ensure that an that checks that the graphics of enemies do not conflict.
    """

    def check_object(self, obj: EnemyObject, level=None, index=0, *args, **kwargs) -> bool:
        """
        Determines if the object should emit a warning for having graphical conflicts.

        Parameters
        ----------
        obj : EnemyObject
            The object to check.
        level : Level
            The level the object is housed inside.
        index : int
            The index into the level enemies obj is housed.

        Returns
        -------
        bool
            If the object should emit a warning.
        """
        if level is None:
            return False
        return any(find_incompatibilities(level.enemies[:index]))

    def get_message(self, obj: ObjectLike) -> str:
        return f"{obj} may have graphical conflicts with other enemies inside the level."


class PydanticEnemyCompatibilityWarning(PydanticWarning):
    """
    A JSON model of a warning that checks that the graphics of enemies do not conflict.
    """
